from django.db import models
import logging
import psycopg2
import psycopg2.extensions

# Create your models here.

#class Call(models.Model):
#    call_start = models.DateTimeField()
#    call_end = models.DateTimeField()
#    caller_id = models.CharField()
#    number_from = models.CharField()
#    number_to = models.CharField()

class LoggingCursor(psycopg2.extensions.cursor):
    def execute(self, sql, args=None):
        logger = logging.getLogger('sql_debug')
        logger.info(self.mogrify(sql, args))

        try:
            psycopg2.extensions.cursor.execute(self, sql, args)
        except Exception, exc:
            logger.error("%s: %s" % (exc.__class__.__name__, exc))
            raise



# -*- coding: utf-8 -*-
# ----------------------------------------------------------------------------
# $Id$
# ----------------------------------------------------------------------------
#
#    Copyright (C) 2008 Caktus Consulting Group, LLC
#
#    This file is part of minibooks.
#
#    minibooks is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Affero General Public License as
#    published by the Free Software Foundation, either version 3 of 
#    the License, or (at your option) any later version.
#    
#    minibooks is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU Affero General Public License for more details.
#    
#    You should have received a copy of the GNU Affero General Public License
#    along with minibooks.  If not, see <http://www.gnu.org/licenses/>.
#

import datetime

from django import forms
from django.db.models import Q
from django.forms.models import inlineformset_factory, BaseInlineFormSet, save_instance
from django.forms.formsets import DELETION_FIELD_NAME

from caktus.decorators import requires_kwarg
from caktus.django.forms import RequestModelForm, RequestForm
from caktus.django.widgets import MooDate

import minibooks.ledger.models as ledger

from crm import models as crm
from timepiece import models as timepiece
from crm.forms import DateInput


class DateForm(RequestForm):
    from_date = forms.DateField(label="From", widget=MooDate(), required=False)
    to_date = forms.DateField(label="To", widget=MooDate(), required=False)
    
    def save(self):
        return (
            self.cleaned_data.get('from_date', ''),
            self.cleaned_data.get('to_date', ''),
        )


class SingleDateForm(RequestForm):
    date = forms.DateField(
        label='Date',
        widget=MooDate(),
        required=False,
    )

    def save(self):
        return self.cleaned_data.get('date', '')


class RepeatPeriodForm(RequestModelForm):
    class Meta:
        model = ledger.RepeatPeriod
        fields = ('active', 'count', 'interval',)
    
    def __init__(self, *args, **kwargs):
        super(RepeatPeriodForm, self).__init__(*args, **kwargs)
    
    def _clean_optional(self, name):
        active = self.cleaned_data.get(name, False)
        value = self.cleaned_data.get(name, '')
        if active and not value:
            raise forms.ValidationError('This field is required.')
        return self.cleaned_data[name]
    
    def clean_count(self):
        return self._clean_optional('count')
    
    def clean_interval(self):
        return self._clean_optional('interval')


class ExchangeTypeForm(RequestModelForm):
    class Meta:
        model = ledger.Exchange
        fields = ('type',)
    
    def __init__(self, *args, **kwargs):
        super(ExchangeTypeForm, self).__init__(*args, **kwargs)
        self.fields['type'] = forms.ModelChoiceField(
            queryset=ledger.ExchangeType.objects.exclude(
                id=self.instance.type.id
            )
        )


class ExchangeForm(RequestModelForm):
    project = forms.ModelChoiceField(
        queryset=timepiece.Project.objects.all(),
        required=False,
    )
    
    class Meta:
        model = ledger.Exchange
        fields = ('business', 'memo', 'date', 'date_due', 'delivered',)
    
    @requires_kwarg('exchange_type')
    def __init__(self, *args, **kwargs):
        self.exchange_type = kwargs.pop('exchange_type')
        super(ExchangeForm, self).__init__(*args, **kwargs)
        
        self.fields.keyOrder = []
        
        if self.request.business:
            self.fields.pop('business')
            business_types = self.request.business.business_types.all()
            if not business_types.filter(can_view_all_projects=True).count():
                self.fields['project'].queryset = \
                  self.request.business.business_projects.all()
        else:
            self.fields.keyOrder.append('business')
            business_types = crm.BusinessType.objects.filter(
                exchange_types=self.exchange_type
            )
            self.fields['business'].queryset = crm.Contact.objects.filter(
                type='business',
                business_types__in=business_types,
            )
            if business_types.count() == 1:
                self.fields['business'].label = business_types[0].name
        
        if self.request.project:
            self.fields.pop('project')
        else:
            self.fields.keyOrder.append('project')
        
        self.fields['memo'].widget = forms.Textarea(
            attrs={'rows': 4, 'cols': 25},
        )
        self.fields['date'].initial = datetime.date.today()
        self.fields['date'].widget = DateInput()
        self.fields['date_due'].widget = DateInput()
        self.fields['date_due'].required = False
        self.fields.keyOrder += ['memo', 'date', 'date_due',]
        
        if self.exchange_type.deliverable:
            self.fields.keyOrder.append('delivered')
        else:
            self.fields.pop('delivered')
        
        debit_or_credit = self.exchange_type.debit_or_credit()
        if self.exchange_type.common_account:
            self.fields[debit_or_credit] = forms.ModelChoiceField(
                queryset=self.exchange_type.common_accounts,
                initial=self.exchange_type.common_account.id,
            )
        if self.instance.id and self.instance.transactions.count() > 0:
            # extract initial data for project and common account 
            # from the first transaction in this exchange
            first_transaction = self.instance.transactions.all()[0]
            if first_transaction.project:
                self.fields['project'].initial = first_transaction.project.id
            value = getattr(first_transaction, debit_or_credit)
            self.fields[debit_or_credit].initial = value.id
            
    def save(self, commit=True):
        instance = super(ExchangeForm, self).save(commit=False)
        if self.request.business:
            instance.business = self.request.business
        instance.type = self.exchange_type
        instance.save(commit)        
        return instance


class BaseTransactionFormSet(BaseInlineFormSet):
    @requires_kwarg('exchange_type')
    def __init__(self, *args, **kwargs):
        self.exchange_type = kwargs.pop('exchange_type')
        super(BaseTransactionFormSet, self).__init__(*args, **kwargs)
    
    def save_existing(self, form, instance, commit=True):
        # ignore amount and quantity from POSTed transaction if reconciled
        exclude = [self._pk_field.name]
        if instance.id and instance.reconciled():
            exclude.extend((
                'amount',
                'quantity',
                'debit_reconciled',
                'credit_reconciled',
            ))
        return save_instance(form, instance, exclude=exclude, commit=commit)
    
    def add_fields(self, form, index):
        super(BaseTransactionFormSet, self).add_fields(form, index)
        form.fields.pop('project')
        
        if index == 0:
            form.fields['date'].initial = datetime.date.today()
        form.fields['date'].widget = DateInput()
        form.fields['memo'].widget = forms.TextInput()
        form.fields['debit_reconciled'].widget = forms.CheckboxInput(
            attrs={'disabled':'disabled'}
        )
        form.fields['credit_reconciled'].widget = forms.CheckboxInput(
            attrs={'disabled':'disabled'}
        )
        
        # disable count and rate fields if transaction is reconciled
        attrs = {}
        if form.instance.id and form.instance.reconciled():
            attrs = {'disabled':'disabled'}
            form.fields['quantity'].required = False
            form.fields['amount'].required = False
            form.fields[DELETION_FIELD_NAME].widget = forms.HiddenInput()
        form.fields['quantity'].widget = forms.TextInput(attrs=attrs)
        form.fields['amount'].widget = forms.TextInput(attrs=attrs)
        
        if self.exchange_type.credit:
            credit_accounts = self.exchange_type.credit_accounts
            form.fields['credit'].queryset = credit_accounts
            form.fields['credit'].initial = self.exchange_type.credit.id
        else:
            form.fields.pop('credit')
        
        if self.exchange_type.debit:
            debit_accounts = self.exchange_type.debit_accounts
            form.fields['debit'].queryset = debit_accounts
            form.fields['debit'].initial = self.exchange_type.debit.id
        else:
            form.fields.pop('debit')


class TransactionForm(forms.ModelForm):
    class Meta:
        model = ledger.Transaction
        
    def clean(self):
        if (self.instance.id
            and self.instance.reconciled()
            and DELETION_FIELD_NAME in self.cleaned_data
            and self.cleaned_data[DELETION_FIELD_NAME]
          ):
            raise forms.ValidationError(
                'Reconciled transactions cannot be deleted'
            )
        return self.cleaned_data


TransactionFormSet = inlineformset_factory(
    ledger.Exchange,
    ledger.Transaction,
    formset=BaseTransactionFormSet,
    form=TransactionForm,
)


class AccountForm(forms.ModelForm):
    """
    Model form for creating and editing ledger accounts.
    """
    
    class Meta:
        model = ledger.Account
        fields = ('name', 'number', 'type',)


class EmailForm(forms.Form):
    to = forms.ChoiceField()
    memo = forms.CharField(max_length=4096, widget=forms.Textarea)
    
    def clean_to(self):
        try:
            return crm.Contact.objects.get(pk=self.cleaned_data['to'])
        except crm.Contact.DoesNotExist:
            raise forms.ValidationError(_(u'That is not a valid contact.'))
        
        return self.cleaned_data['to']
    
    def __init__(self, *args, **kwargs):
        search = None
        
        business = kwargs.pop('business', None)
        project = kwargs.pop('project', None)
        
        if project:
            # ignore business contacts if a project is set
            search = Q(contact_projects=project)
        elif business:
            search = Q(contacts=business)
        
        super(EmailForm, self).__init__(*args, **kwargs)
        
        if search:
            self.fields['to'].choices = []
            for contact in crm.Contact.objects.filter(search):
                choice = (contact.id, "%s (%s)" % (contact.get_full_name(), contact.email))
                self.fields['to'].choices.append(choice)


class SearchForm(forms.Form):
    search = forms.CharField(required=False)


class ReconcileTransactionForm(forms.Form):
    ACCOUNTS = (
        ('debit', 'Debit',),
        ('credit', 'Credit',),
    )
    account = forms.ChoiceField(choices=ACCOUNTS)
    transaction = forms.ModelChoiceField(
        queryset=ledger.Transaction.objects.all(),
    )
    to_date = forms.DateField(required=False)


class TransferForm(forms.ModelForm):
    class Meta:
        model = ledger.Transaction
        fields = ('credit', 'debit', 'amount', 'date',)
    
    def __init__(self, *args, **kwargs):
        super(TransferForm, self).__init__(*args, **kwargs)
        self.fields.keyOrder = ('credit', 'debit', 'date', 'amount',)
        self.fields['debit'].label = 'To'
        self.fields['debit'].queryset = ledger.Account.objects.all()
        self.fields['credit'].label = 'From'
        self.fields['credit'].queryset = ledger.Account.objects.filter(
            type__in=('asset', 'liability')
        )
        self.fields['date'].widget = DateInput()
        
        if self.instance.id and self.instance.reconciled():
            self.fields['amount'].required = False
            self.fields['amount'].widget.attrs = {'disabled': 'disabled'}
    
    def clean_debit(self):
        debit = self.cleaned_data['debit']
        credit = self.cleaned_data['credit']
        if debit.type != credit.type or debit.id == credit.id:
            raise forms.ValidationError('Select a valid choice. That choice is not one of the available choices.')
        return debit
    
    def save(self):
        # ignore POSTed amount field if transaction is reconciled
        if self.instance.id and self.instance.reconciled():
            self.cleaned_data.pop('amount')
        instance = super(TransferForm, self).save(commit=False)
        instance.memo = 'Transfer from %s to %s' % (
            instance.credit, 
            instance.debit,
        )
        instance.quantity = 1
        instance.save()
        return instance


class GeneralEntryForm(RequestModelForm):
    class Meta:
        model = ledger.Transaction
        fields = ('date', 'debit', 'credit', 'memo', 'amount',)
    
    def __init__(self, *args, **kwargs):
        super(GeneralEntryForm, self).__init__(*args, **kwargs)
        self.fields.keyOrder = (
            'credit',
            'debit',
            'date',
            'amount',
            'memo',
        )
        self.fields['date'].initial = datetime.date.today()
        self.fields['date'].widget = DateInput()
        if self.instance.id and self.instance.reconciled():
            self.fields['amount'].required = False
            self.fields['amount'].widget.attrs = {'disabled': 'disabled'}
    
    def save(self):
        # ignore POSTed amount field if transaction is reconciled
        if self.instance.id and self.instance.reconciled():
            self.cleaned_data.pop('amount')
        instance = super(GeneralEntryForm, self).save(commit=False)
        instance.quantity = 1
        instance.save()
        return instance



# -*- coding: utf-8 -*-
# ----------------------------------------------------------------------------
# $Id$
# ----------------------------------------------------------------------------
#
#    Copyright (C) 2008 Caktus Consulting Group, LLC
#
#    This file is part of minibooks.
#
#    minibooks is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Affero General Public License as
#    published by the Free Software Foundation, either version 3 of 
#    the License, or (at your option) any later version.
#    
#    minibooks is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU Affero General Public License for more details.
#    
#    You should have received a copy of the GNU Affero General Public License
#    along with minibooks.  If not, see <http://www.gnu.org/licenses/>.
#


from django import template

register = template.Library()

@register.filter(name='index')
def index(list, arg):
    return list[arg]



# -*- coding: utf-8 -*-
# ----------------------------------------------------------------------------
# $Id$
# ----------------------------------------------------------------------------
#
#    Copyright (C) 2008 Caktus Consulting Group, LLC
#
#    This file is part of minibooks.
#
#    minibooks is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Affero General Public License as
#    published by the Free Software Foundation, either version 3 of 
#    the License, or (at your option) any later version.
#    
#    minibooks is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU Affero General Public License for more details.
#    
#    You should have received a copy of the GNU Affero General Public License
#    along with minibooks.  If not, see <http://www.gnu.org/licenses/>.
#


from django.core.mail import EmailMessage
from django.template import RequestContext, Context, loader


def send_caktus_email(exchange_slug=None, exchange_label=None, recipients=None, subject=None, context=None, attachment=None):
    """
        send_caktus_email('invoice', context, recipients, attachment)
    """
    
    if not context:
        context = {}
    
    # prepare email message
    template = loader.get_template("books/ledger/exchange/email/%s.txt" % exchange_slug)
    template_context = Context(context)
    
    # generate email message
    body = template.render(template_context)
    
    if subject is None:
        subject = "Caktus %s" % exchange_label
    
    sender = "team@caktusgroup.com"
    
    try:
        email = EmailMessage(subject, body, sender, recipients, [sender])
    
        if attachment:
            email.attach('%s.pdf' % exchange_slug, attachment)
            
        email.send()
        
        return True
        
    except Exception, err:
        return False

def send_exchange_email(exchange_type=None, recipients=None, subject=None, context=None, attachment=None):
    return send_caktus_email(exchange_slug="exchange_email", exchange_label=exchange_type.label, recipients=recipients, subject=subject, context=context, attachment=attachment)

def send_project_report_email(recipients=None, subject=None, context=None, attachment=None):
    return send_caktus_email(exchange_slug="project_report", exchange_label="Project Report", recipients=recipients, subject=subject, context=context, attachment=attachment)



# -*- coding: utf-8 -*-
# ----------------------------------------------------------------------------
# $Id$
# ----------------------------------------------------------------------------
#
#    Copyright (C) 2008 Caktus Consulting Group, LLC
#
#    This file is part of minibooks.
#
#    minibooks is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Affero General Public License as
#    published by the Free Software Foundation, either version 3 of 
#    the License, or (at your option) any later version.
#    
#    minibooks is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU Affero General Public License for more details.
#    
#    You should have received a copy of the GNU Affero General Public License
#    along with minibooks.  If not, see <http://www.gnu.org/licenses/>.
#


import datetime

from django.conf import settings

from minibooks.ledger.models import Account
from minibooks.ledger.pdf_latex import render_to_pdf_string


def generate_exchange_pdf(exchange):
    """ invoices, receipts, etc """
    
    try:
        location = exchange.business.locations.all()[0]
    except IndexError:
        location = None
    
    if location:
        try:
            client_address = location.addresses.all()[0]
        except IndexError:
            client_address = None
    else:
        client_address = None
        
    context = {
        'exchange': exchange,
        'letter_title': '%s \#%d' % (exchange.type, exchange.id),
        'letter_date': exchange.date,
        'client': exchange.business,
        'client_address': client_address,
        'show_project_balance': exchange.type.slug in ('invoice', 'receipt'),
        'MEDIA_ROOT': settings.MEDIA_ROOT,
    }
    
    return render_to_pdf_string('books/ledger/exchange/view.tex', context)


def generate_project_report_pdf(project):
    """ credits and debits of a Project """
    
    ar = Account.objects.get(number=1200)
    context = {
        'ar'      : ar,
        'credits' : ar.credits.filter(project=project.id),
        'debits'  : ar.debits.filter(project=project.id),
        'client'  : project.business,
        'project' : project,
        'letter_title': 'Project Report',
        'letter_date': datetime.datetime.now(),
        'MEDIA_ROOT': settings.MEDIA_ROOT,
    }
    
    credit_total = ar.credit_total_for_project(project.id)
    debit_total = ar.debit_total_for_project(project.id)
    
    if not credit_total: credit_total = 0
    if not debit_total:   debit_total = 0
    outstanding_balance = debit_total - credit_total

    context.update( { 
                   'credit_total'        : credit_total,
               'debit_total'         : debit_total,
               'outstanding_balance' : outstanding_balance
               })
    
    return render_to_pdf_string('books/report/project.tex', context)



from trac_tickets.alchemy import *
from pylab import *

def prep_figure(title, x_label, y_label):
    fig = Figure()
    canvas = FigureCanvas(fig)
    ax = fig.add_subplot(111)
    ax.set_title(title)
    ax.set_xlabel(x_label)
    ax.set_ylabel(y_label)
    fig.autofmt_xdate(rotation=45)
    return (ax, canvas)

def render_canvas(canvas, format):
    canvas.draw()
    size = canvas.get_renderer().get_canvas_width_height()
    buf = canvas.tostring_rgb()
    im = Image.fromstring('RGB', size, buf, 'raw', 'RGB', 0, 1)
    imdata = StringIO()
    im.save(imdata, format=format)
    return imdata.getvalue()

def graph_hours_sum(authors, format='JPEG'):
    """ Returns data for a graph of hour sums"""
    ax, canvas = prep_figure(title='Developer Hour Sums',
                                 x_label='time',
                                                     y_label='hours commited')
    styles = ['g-','r-','b-', 'k-', 'c-', 'm-', 'y-' ]
    for i, author in enumerate(authors):
        dates = []
        hours = []
        sums = []
        sum = 0
        for day in session.query(DayAuthorHoursSummary).filter("author='%s'" % author).order_by('date').all():
            sum = sum + float(day.hours)
            sums.append(sum)
            dates.append(date2num(day.date))
            hours.append(float(day.hours))
        hours = hours[1:-1]
        dates = dates[1:-1]
        sums = sums[1:-1]
        if len(dates) != 0:
            ax.plot_date(dates, sums, styles[i % len(styles)], label=author)
    session.close()
    return render_canvas(canvas, format)


def graph_developer_hour_commits(author, format='JPEG'):
    """ Returns a scatter plot of a developer's commited hours"""
    ax, canvas = prep_figure(
        title='Commit Hours: %s' % author,
        x_label='time',
        y_label='hours commited',
    )
    dates = []
    hours = []
    for day in session.query(DayAuthorHoursSummary).filter(
      "author='%s'" % author).order_by('date').all():
        dates.append(date2num(day.date))
        hours.append(float(day.hours))
    session.close()
    if len(dates) != 0:
        ax.plot_date(dates, hours, 'go', label=author)
    return render_canvas(canvas, format)


def graph_account_sum(account, format='JPEG'):
    ax, canvas = prep_figure(
        title='%s' % account.name,
        x_label='time',
        y_label='account input sum',
    )
    sums = []
    dates = []
    for transaction in account.get_table()[2]:
        sums.append(transaction[-1])
        dates.append(date2num(transaction[0]))
    if len(dates) != 0:
        ax.plot_date(dates, sums, '-')
    return render_canvas(canvas, format)



import psycopg2
import psycopg2.extensions

from caktus.decorators import requires_kwarg

class CaktusCursor(psycopg2.extensions.cursor):
    def execute(self, sql, args=None):
        print sql
        psycopg2.extensions.cursor.execute(self, sql, args)

class CaktusTrac(object):
    dsn = None
    client_encoding = 'UTF8'
    dbname, user, password, host = None, None, None, None
    
    @requires_kwarg('dbname')
    def __init__(self, *args, **kwargs):
        self.dbname = kwargs.pop('dbname')
        self.user = kwargs.pop('user', None)
        self.password = kwargs.pop('password', None)
        self.host = kwargs.pop('host', None)
        
        self._connect()
    
    def _connect(self):
        """
        Establish a connection to trac database
        """
        
        # create DSN string
        dsn_options = []
        for option in ('dbname', 'user', 'password', 'host'):
            value = getattr(self, option, None)
            if value is not None:
                dsn_options.append("%s='%s'" % (option, value))
        
        self.dsn = " ".join(dsn_options)
        self.connection = psycopg2.connect(self.dsn)
        self.connection.set_client_encoding(self.client_encoding)
        self.cursor = self.connection.cursor(cursor_factory=CaktusCursor)
    
    def _install_views(self):
        """
        Create views to simplify access to trac database model
        """
        
        # get a list of trac schemas (one for each trac project)
        self.cursor.execute(
        """
        SELECT
            DISTINCT(table_schema)
        FROM 
            information_schema.tables
        WHERE
            table_schema NOT IN ('information_schema', 'pg_catalog', 'public')
        """)
        
        for schema in self.cursor.fetchall():
            schema_name = schema[0]
            
            try:
                self.cursor.execute("DROP VIEW _%s_totalhours CASCADE" % schema_name)
            except psycopg2.ProgrammingError:
                # view does not exist
                self.connection.rollback()
            
            self.cursor.execute("""
            CREATE VIEW _%s_totalhours AS
            SELECT
                change.ticket,
                ticket.milestone,
                change.author,
                ROUND(CASE 
                        WHEN newvalue = '' OR newvalue is NULL THEN 0
                        ELSE newvalue::float 
                    END::numeric, 2) AS value,
                value as billable
            FROM %s.ticket_change change, %s.ticket_custom custom, %s.ticket ticket
            WHERE change.ticket = custom.ticket AND
                change.ticket = ticket.id AND
                custom.name = 'billable' AND
                change.field = 'totalhours' AND
                change.time = (SELECT max(change2.time) 
                        FROM %s.ticket_change change2 
                        WHERE change2.ticket = change.ticket AND 
                            change2.field='totalhours')
            GROUP BY change.author, change.ticket, ticket.milestone, custom.value, change.newvalue
            ORDER BY change.ticket;
            """ % (schema_name, schema_name, schema_name, schema_name, schema_name))
            
            try:
                self.cursor.execute("DROP VIEW _%s_hours_with_time CASCADE" % schema_name)
            except psycopg2.ProgrammingError:
                self.connection.rollback()
            
            self.cursor.execute("""
            CREATE VIEW _%s_hours_with_time AS
            SELECT     
                change.ticket,
                to_timestamp(change.time) AS time,
                ticket.milestone,
                change.author,
                ROUND(CASE 
                    WHEN newvalue = '' OR newvalue is NULL THEN 0
                    ELSE newvalue::float 
                    END::numeric, 2) AS value,
                value as billable
            FROM %s.ticket_change change, %s.ticket_custom custom, %s.ticket ticket
            WHERE change.ticket = custom.ticket AND
                change.ticket = ticket.id AND
                custom.name = 'billable' AND
                change.field = 'hours'
            GROUP BY change.author, change.ticket, ticket.milestone, custom.value, change.time, change.newvalue
            ORDER BY change.ticket;
            """ % (schema_name, schema_name, schema_name, schema_name))
            
            try:
                self.cursor.execute("DROP VIEW _%s_hours CASCADE" % schema_name)
            except psycopg2.ProgrammingError:
                self.connection.rollback()
            
            return
            
            self.cursor.execute("""
            CREATE VIEW _%s_hours AS
            SELECT
                ticket,
                milestone,
                author,
                SUM(value) AS value,
                billable
            FROM _%s_hours_with_time
            GROUP BY value, ticket, milestone, author, billable;
            """ % (schema_name, schema_name))
            
            combined_view_sql.append("""
            SELECT '%s' as project,
                hours.ticket, 
                hours.milestone,
                hours.author,
                hours.value AS hours,
                totalhours.value AS totalhours,
                hours.billable
            FROM _%s_totalhours totalhours, _%s_hours hours
            WHERE totalhours.ticket = hours.ticket
            """ % (schema_name, schema_name, schema_name))
            
            combined_view_sql_with_time.append("""
            SELECT
                '%s' as project,
                ticket, 
                time,
                milestone,
                author,
                value AS hours,
                billable
            FROM
                _%s_hours_with_time
            """ % (schema_name, schema_name))
            
        self.cursor.execute("""
        CREATE VIEW public.combined_hours AS
        %s;
        """ % '\nUNION ALL\n'.join(combined_view_sql))
        
        self.cursor.execute("""
        CREATE VIEW public.combined_hours_with_time AS
        %s;
        """ % '\nUNION ALL\n'.join(combined_view_sql_with_time))
    



# -*- coding: utf-8 -*-
# ----------------------------------------------------------------------------
# $Id$
# ----------------------------------------------------------------------------
#
#    Copyright (C) 2008 Caktus Consulting Group, LLC
#
#    This file is part of minibooks.
#
#    minibooks is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Affero General Public License as
#    published by the Free Software Foundation, either version 3 of 
#    the License, or (at your option) any later version.
#    
#    minibooks is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU Affero General Public License for more details.
#    
#    You should have received a copy of the GNU Affero General Public License
#    along with minibooks.  If not, see <http://www.gnu.org/licenses/>.
#


import datetime
import re

try:
    from dateutil.relativedelta import relativedelta
except ImportError:
    raise ImportError('minibooks was unable to import the dateutil library. Please confirm it is installed and available on your current Python path.')

from django.db import models
from django.db.models import Q
from django.contrib.auth.models import User, Group, Permission
from django.contrib.contenttypes.models import ContentType
from django.contrib.sites.models import Site
from django.conf import settings
from django.core.urlresolvers import reverse

from caktus.django.db.util import slugify_uniquely
from caktus.django.db.util import query_to_table

from crm import models as crm

from timepiece import models as timepiece

class AccountManager(models.Manager):
    def get_query_set(self):
        debit_total_sql = """
        SELECT COALESCE(SUM(ROUND(amount * quantity, 2)),0.0) 
        FROM ledger_transaction 
        WHERE debit_id = ledger_account.id
        """
        
        credit_total_sql = """
        SELECT COALESCE(SUM(ROUND(amount * quantity, 2)),0.0)
        FROM ledger_transaction
        WHERE credit_id = ledger_account.id
        """
        
    #    balance_sql = """
    #    COALESCE((%s) - (%s), 0.0)
    #    """ % (debit_total_sql, credit_total_sql)
        
        return super(AccountManager, self).get_query_set().extra(
            select={
                'debit_total': debit_total_sql,
                'credit_total': credit_total_sql,
    #            'balance': balance_sql,
            }
        )


class Account(models.Model):
    """
    Accounts for use in your business chart of accounts
    
    There will normally be far more expense accounts than any other type of account.
    
    The information which you collect in your income accounts will be used to prepare your profit and loss statements periodically.
    Profit and loss statement = income and expense statement
    
    phone - Telephone expenses
    google - Advertising expenses
    celito - 
    business meals and lodging
    business insurance
    charitable contributions
    auto expenses
    server expenses
    
    Asset and liability accounts are collectively referred to as balance sheet accounts
        Accounts receivable (current asset)
        Bank checking account (current asset)
        
        Accounts payable (current debt)
        Equipement (fixed asset)
    """
    
    ACCOUNT_TYPES = (
        ('income', 'Income' ),
        ('expense', 'Expense' ),
        ('asset','Asset'),
        ('liability', 'Liability'),
        ('equity', 'Equity'),
    )
    
    name = models.CharField(max_length=255, unique=True)
    number = models.IntegerField(unique=True)
    type = models.CharField(max_length=15, choices=ACCOUNT_TYPES)
    
    objects = AccountManager()

    def balance(self):
        if self.type in ('asset', 'expense', 'equity',):
            return self.debit_total - self.credit_total
        else:
            return self.credit_total - self.debit_total
    
    def _difference(self, debit_amt, credit_amt):
        if self.type in ('asset', 'expense', 'equity',):
            return debit_amt - credit_amt
        else:
            return credit_amt - debit_amt
        
    def _total(self, type=None, reconciled=None, **kwargs):
        if type in ('credit', 'debit'):
            kwargs[type] = self
            if reconciled is not None:
                kwargs['%s_reconciled' % type] = reconciled
        else:
            raise Exception('Invalid type %s' % type)
        
        return Transaction.sum(**kwargs)
    
    def reconciled_balance(self, to_date=None):
        debit_total = self._total('debit', reconciled=True, to_date=to_date)
        credit_total = self._total('credit', reconciled=True, to_date=to_date)
        return self._difference(debit_total, credit_total)
    
    def debit_total_for_project(self, project_id=None):
        return self._total('debit', project_id=project_id)
        
    def credit_total_for_project(self, project_id=None):
        return self._total('credit', project_id=project_id)    
    
    def total_for_date_range(self, from_date=None, to_date=None):
        debit_total = self._total('debit', from_date=from_date, to_date=to_date)
        credit_total = self._total('credit', from_date=from_date, to_date=to_date)
        return self._difference(debit_total, credit_total)
    
    def get_transactions(self, from_date=None, to_date=None):
        if self.type in ('asset', 'expense', 'equity',):
            balance_order = ('debit', 'credit',)
        else:
            balance_order = ('credit', 'debit',)
        
        transactions = Transaction.objects.filter(
            Q(credit=self) | Q(debit=self)
        ).extra(
            select={
                'balance': 'ledger_account_balance.%s_balance - ledger_account_balance.%s_balance' % balance_order,
            },
            tables=[
            #    'running_balance(%s) ledger_account_balance' % self.id,
                 'ledger_account_balance',
            ],
            where=[
                'ledger_account_balance.account_id = %s',
                'ledger_account_balance.transaction_id = ledger_transaction.id',
            ],
            params=[
                self.id,
            ]
        # this order MUST match the order in ledger_account_ledger
        ).order_by(
            'date',
            'exchange__id',
            'id',
        ).select_related(
            'exchange__business', 
            'credit', 
            'debit', 
            'exchange', 
            'exchange__type',
        )
        
        if from_date:
            transactions = transactions.filter(date__gte=from_date)
            
        if to_date:
            transactions = transactions.filter(date__lte=to_date)
            
        return transactions
    
    def get_table(self, from_date=None, to_date=None):
        sql = "SELECT date, business, memo, debit, credit, total FROM ledger_account_ledger WHERE account = %s"
        args = [self.id,]
        
        if from_date:
            sql += " AND date >= %s"
            args.append(from_date)
            
        if to_date:
            sql += " AND date <= %s"
            args.append(to_date)
        
        return query_to_table(sql, args)
    
    @classmethod
    def profit_loss(cls, from_date=None, to_date=None):
        total_profits = 0
        total_losses = 0
        profit_accounts = cls.objects.filter(
            type='income',
        ).order_by('name')
        loss_accounts = cls.objects.filter(
            type='expense',
        ).order_by('name')
        
        for account in profit_accounts:
            account.date_range_total = account.total_for_date_range(
                from_date=from_date,
                to_date=to_date,
            )
            total_profits += account.date_range_total
        
        for account in loss_accounts:
            account.date_range_total = account.total_for_date_range(
                from_date=from_date,
                to_date=to_date,
            )
            total_losses += account.date_range_total
        
        return {
            'profit_accounts': profit_accounts,
            'loss_accounts': loss_accounts,
            'total_profits': total_profits,
            'total_losses': total_losses,
            'net_income': total_profits - total_losses,
        }
    
    @classmethod
    def sum_by_type(cls, account_type, from_date, to_date):
        sql = ''
        if account_type in ('asset', 'expense', 'equity',):
            sql += 'SELECT SUM(debit_sum) - SUM(credit_sum) AS sum'
        else:
            sql += 'SELECT SUM(credit_sum) - SUM(debit_sum) AS sum'
        args = []
        args.append(from_date)
        args.append(to_date)
        args.append(from_date)
        args.append(to_date)
        args.append(account_type)
        sql += """
        FROM
            (
                SELECT
                    (SELECT COALESCE(SUM(ROUND(amount * quantity, 2)), 0.0)
                        FROM
                            ledger_transaction
                        WHERE
                            ledger_transaction.credit_id = ledger_account.id
                            AND ledger_transaction.date >= %s
                            AND ledger_transaction.date < %s
                    ) AS credit_sum,
                    (SELECT COALESCE(SUM(ROUND(amount * quantity, 2)), 0.0)
                        FROM
                            ledger_transaction
                        WHERE
                            ledger_transaction.debit_id = ledger_account.id
                            AND ledger_transaction.date >= %s
                            AND ledger_transaction.date < %s
                    ) AS debit_sum
                FROM
                    ledger_account
                WHERE
                    ledger_account.type = %s
            ) AS debit_and_credit_sum
        """
        from django.db import connection
        cursor = connection.cursor()
        cursor.execute(sql, args)
        row = cursor.fetchone()
        return row[0]
        
    class Meta:
        ordering = ('type', 'name',)
        permissions = (
            ('view_account', 'Can view account'),
            ('view_account_reports', 'Can view account reports'),
            ('transfer_funds', 'Can transfer funds')
        )
        
    def __unicode__(self):
        return "%i - %s (%s)" % (self.number, self.name, self.type)


class ExchangeTypeManager(models.Manager):
    def get_query_set(self):
        return super(ExchangeTypeManager, self).get_query_set().extra(
            select={
                'exchange_count': 'SELECT COUNT(*) FROM ledger_exchange WHERE type_id = ledger_exchangetype.id',
            }
        )
    pass


class ExchangeType(models.Model):
    """
    Links exchanges to default accounts
    """
    
    label = models.CharField(max_length=100, unique=True)
    slug = models.CharField(max_length=100, unique=True)
    debit = models.ForeignKey(
        Account,
        null=True,
        blank=True,
        related_name='debit_exchange_types',
    )
    credit = models.ForeignKey(
        Account,
        null=True,
        blank=True,
        related_name='credit_exchange_types',
    )
    common_account = models.ForeignKey(
        Account,
        null=True,
        blank=True,
        related_name='common_account_exchange_types',
    )
    deliverable = models.BooleanField(default=False)
    
    business_types = models.ManyToManyField(
        crm.BusinessType,
        related_name='exchange_types',
    )
    
    objects = ExchangeTypeManager()
    
    class Meta:
        ordering = ('label',)
    
    def save(self, *args, **kwargs):
        queryset = ExchangeType.objects.all()
        if self.id:
            queryset = queryset.exclude(id__exact=self.id)
        self.slug = slugify_uniquely(self.label, queryset, 'slug')
        super(ExchangeType, self).save(*args, **kwargs)
    
    def debit_or_credit(self):
        if self.common_account:
            if self.debit is None:
                return 'debit'
            elif self.credit is None:
                return 'credit'
            else:
                raise ValueError('common_account must take the place of the debit or credit account in ExchangeType')
        return None

    def _get_credit_accounts(self, base_account):
        # TODO make this stuff part of the data model
        if base_account and self.slug == 'invoice':
            qs = Account.objects.filter(Q(type='income') | Q(type='expense'))
        elif base_account and self.slug == 'purchase':
            qs = Account.objects.filter(Q(type='liability') | Q(type='asset'))
        elif base_account:
            qs = Account.objects.filter(type=base_account.type)
        else:
            qs = Account.objects.none()
        return qs
    credit_accounts = property(
        lambda self: self._get_credit_accounts(self.credit)
    )

    def _get_debit_accounts(self, base_account):
        if base_account and self.slug == 'purchase-credit':
            qs = Account.objects.filter(Q(type='liability') | Q(type='asset'))
        elif base_account:
            qs = Account.objects.filter(type=base_account.type)
        else:
            qs = Account.objects.none()
        return qs
    debit_accounts = property(
        lambda self: self._get_debit_accounts(self.debit)
    )
    
    def _get_common_accounts(self):
        if self.debit:
            qs = self._get_credit_accounts(self.common_account)
        elif self.credit:
            qs = self._get_debit_accounts(self.common_account)
        else:
            qs = Account.objects.none()
        return qs
    common_accounts = property(_get_common_accounts)
    
    def __unicode__(self):
        return self.label


class ExchangeManager(models.Manager):
    def get_query_set(self):
        return super(ExchangeManager, self).get_query_set().extra(
            select={
                'total': 'SELECT SUM(ROUND(amount * quantity, 2)) FROM ledger_transaction WHERE exchange_id = ledger_exchange.id',
                'reconciled': 'SELECT bool_or(t.debit_reconciled OR t.credit_reconciled) FROM ledger_transaction t WHERE exchange_id = ledger_exchange.id',
            }
        )


class RepeatPeriod(models.Model):
    INTERVAL_CHOICES = (
        ('day', 'Day(s)'),
        ('week', 'Week(s)'),
        ('month', 'Month(s)'),
        ('year', 'Year(s)'),
    )
    active = models.BooleanField(default=False)
    count = models.PositiveSmallIntegerField(null=True, blank=True, choices=[(x,x) for x in range(1,32)])
    interval = models.CharField(null=True, blank=True, max_length=10, choices=INTERVAL_CHOICES)

    def delta(self):
        return relativedelta(**{str(self.interval + 's'): self.count})


class Exchange(models.Model):
    """ Collection of specific transactions """
    
    business = models.ForeignKey(crm.Contact, related_name="exchanges")
    type = models.ForeignKey(ExchangeType, related_name='exchanges')
    memo = models.TextField(null=True, blank=True)
    date = models.DateField()
    date_due = models.DateField(null=True, blank=True)
    delivered = models.BooleanField(default=False)
    
    repeat_period = models.ForeignKey(RepeatPeriod, null=True, blank=True, related_name='exchanges')
    
    objects = ExchangeManager()
    
    def previous_project_balance(self):
        if self.transactions.all().count() > 0:
            return self.transactions.all()[0].previous_project_balance()
        
    def new_project_balance(self):
        if self.type.slug == 'receipt':
            return self.previous_project_balance() - self.total
        elif self.type.slug == 'invoice':
            return self.previous_project_balance() + self.total
    
    def get_common_account(self):
        if self.transactions.all().count() > 0 and self.type.debit_or_credit():
            return getattr(self.transactions.all()[0], self.type.debit_or_credit())
        else:
            return None
        
    def get_project(self):
        if self.transactions.all().count() > 0:
            return self.transactions.all()[0].project
        else:
            return None
    
    def subtotal(self):
        return self.total
    
    @classmethod
    def get_user_payment_hours(cls):
        from trac_tickets.models import TracConnection
        projects = crm.Project.objects.order_by('name')
        try:
            trac = TracConnection()
    
            paycheck = ExchangeType.objects.get(label='Paycheck')
            now = datetime.datetime.now()
            tomorrow = now + datetime.timedelta(days=1)
            previous_pay_date_end = Exchange.objects.filter(type=paycheck)[0].date - datetime.timedelta(days=1)
            previous_pay_date_start = Exchange.objects.filter(type=paycheck,date__lt=previous_pay_date_end)[0].date - datetime.timedelta(days=1)
    
            curr_paychecks = trac.getUserTotalHours( projects, min_date=previous_pay_date_end, max_date=tomorrow, hash_prefix='curr_paychecks')
            prev_paychecks = trac.getUserTotalHours( projects, min_date=previous_pay_date_start, max_date=previous_pay_date_end, hash_prefix='prev_paychecks')
            paychecks = {}
            for key in curr_paychecks.keys():
                paychecks[key] = curr_paychecks[key]
                paychecks[key].update( prev_paychecks[key] )
    
        except Exception, e:
            raise

        return paychecks
        
    class Meta:
        ordering = ('-date', '-id', 'type',)
        permissions = (
            ('view_exchange', 'Can view exchange'),
            ('email_exchange', 'Can email exchange'),
        )

    def __unicode__(self):
        if self.id:
            return "%s: %s" % (self.type, self.memo)
        else:
            return unicode(self.memo)


class TransactionManager(models.Manager):
    def get_query_set(self):
        return super(TransactionManager, self).get_query_set().extra(
            select={
                'total': 'ROUND(amount * quantity, 2)'
            }
        )


class Transaction(models.Model):
    date = models.DateField()
    debit = models.ForeignKey(Account, related_name='debits')
    credit = models.ForeignKey(Account, related_name='credits')
    project = models.ForeignKey(timepiece.Project, null=True, related_name='transactions')
    exchange = models.ForeignKey(Exchange, null=True, related_name='transactions')
    
    memo = models.TextField()
    amount = models.DecimalField(max_digits=10, decimal_places=2)
    quantity = models.DecimalField(max_digits=10, decimal_places=2)
    debit_reconciled = models.BooleanField()
    credit_reconciled = models.BooleanField()
    
    objects = TransactionManager()
    
    def previous_project_balance(self):
        billed = Transaction.sum(
            project_id=self.project.id,
            before=self,
            exchange_type=ExchangeType.objects.get(slug='invoice'),
        )
        paid = Transaction.sum(
            project_id=self.project.id,
            before=self,
            exchange_type=ExchangeType.objects.get(slug='receipt'),
        )
        return billed - paid
    
    def new_project_balance(self):
        return self.previous_project_balance() + self.total
    
    @classmethod
    def sum(cls, 
        debit=None, 
        credit=None, 
        project_id=None, 
        debit_reconciled=None, 
        credit_reconciled=None, 
        from_date=None, 
        to_date=None,
        before=None,
        exchange_type=None,
      ):
        args = []
        where = []
        join = []
        
        if debit:
            where.append("ledger_transaction.debit_id = %s")
            args.append(debit.id)
        if credit:
            where.append("ledger_transaction.credit_id = %s")
            args.append(credit.id)
        if project_id:
            where.append("ledger_transaction.project_id = %s")
            args.append(project_id)
        if from_date:
            where.append("ledger_transaction.date >= %s")
            args.append(from_date)
        if to_date:
            where.append("ledger_transaction.date <= %s")
            args.append(to_date)
        if debit_reconciled is not None:
            where.append("ledger_transaction.debit_reconciled = %s")
            args.append(debit_reconciled)
        if credit_reconciled is not None:
            where.append("ledger_transaction.credit_reconciled = %s")
            args.append(credit_reconciled)
        if before:
            where.append("""
            (ledger_exchange.date < %s OR 
            (ledger_exchange.date = %s AND ledger_exchange.id < %s) OR
            (ledger_exchange.date = %s AND ledger_exchange.id = %s AND ledger_transaction.date < %s) OR
            (ledger_exchange.date = %s AND ledger_exchange.id = %s AND ledger_transaction.date = %s AND ledger_transaction.id < %s))
            """)
            args.append(before.exchange.date)
            args.append(before.exchange.date)
            args.append(before.exchange.id)
            args.append(before.exchange.date)
            args.append(before.exchange.id)
            args.append(before.date)
            args.append(before.exchange.date)
            args.append(before.exchange.id)
            args.append(before.date)
            args.append(before.id)
        if exchange_type:
            join.append('JOIN ledger_exchange ON ledger_exchange.id = ledger_transaction.exchange_id')
            where.append('ledger_exchange.type_id = %s')
            args.append(exchange_type.id)
            
        sql = "SELECT COALESCE(SUM(ROUND(amount * quantity, 2)), 0.0) FROM ledger_transaction"
        if join:
            sql += ' ' + ' '.join(join)
        if where:
            sql += " WHERE %s" % ' AND '.join(where)
            
        from django.db import connection
        cursor = connection.cursor()
        cursor.execute(sql, args)
        row = cursor.fetchone()
        return row[0]
    
    def reconciled(self):
        return (self.debit_reconciled or self.credit_reconciled)
    
    def __unicode__(self):
        return '%s: %s (%s)' % (self.date, self.memo, self.amount)
        
    class Meta:
        ordering = ('exchange__date', 'exchange__id', 'date', 'id',)


class ProjectManager(models.Manager):
    def get_query_set(self):
        invoice_sql = """
SELECT COALESCE(SUM(ROUND(amount * quantity, 2)), 0.0) 
FROM ledger_transaction t
INNER JOIN ledger_exchange e 
  ON (t.exchange_id=e.id) 
INNER JOIN ledger_exchangetype et 
  ON (e.type_id=et.id) 
WHERE t.project_id=crm_project.id
  AND et.slug='invoice'
        """
        receipt_sql = """
SELECT COALESCE(SUM(ROUND(amount * quantity, 2)), 0.0) 
FROM ledger_transaction t
INNER JOIN ledger_exchange e 
  ON (t.exchange_id=e.id) 
INNER JOIN ledger_exchangetype et 
  ON (e.type_id=et.id) 
WHERE t.project_id=crm_project.id
  AND et.slug='receipt'
        """
        return super(ProjectManager, self).get_query_set().extra(
            select={
                'balance': '(%s) - (%s)' % (invoice_sql, receipt_sql),
            }
        )
#crm.Project.objects = ProjectManager()


def install():
    group, created = Group.objects.get_or_create(name='Ledger Admin')
    if created:
        perms = Permission.objects.filter(
            content_type__in=ContentType.objects.filter(app_label='ledger')
        )
        for perm in perms:
            group.permissions.add(perm)
    
    if Account.objects.count() > 0 or ExchangeType.objects.count() > 0:
        return
    
    account_list = (
        (1060, 'Checking', 'asset'),
        (1200, 'Accounts Receivable', 'asset'),
        (1300, 'Owner Draws', 'asset'),
        (2100, 'Accounts Payable', 'liability'),
        (2200, 'Credit Card', 'liability'),
        (4020, 'Consulting', 'income'),
        (5020, 'General Expenses', 'expense'),
        (5615, 'Advertising & Promotions', 'expense'),
        (5620, 'Telephone', 'expense'),
        (5685, 'Insurance', 'expense'),
        (5695, 'Internet', 'expense'),
    )    
    
    exchange_type_list = (
        ('Invoice',  None, 4020, 1200),
        ('Receipt',  1060, None, 1200),
        ('Order',    5020, None, 2100),
        ('Payment',  None, 1060, 2100),
        ('Paycheck',  1060, None, 1300),
        ('Purchase', 5020, None, 1060),
        ('Purchase Credit', None, 5020, 2200),
        ('Credit Card Payment', None, 1060, 2200),
    )
    
    for number, name, type in account_list:
        Account(number=number, name=name, type=type).save()
    
    for label, debit, credit, common_account in exchange_type_list:
        tt = ExchangeType(label=label)
        if debit is not None:
            tt.debit = Account.objects.get(number=debit)
        if credit is not None:
            tt.credit = Account.objects.get(number=credit)
        if common_account is not None:
            tt.common_account = Account.objects.get(number=common_account)
        tt.save()
    
    if crm.BusinessType.objects.count() > 0:
        return
    
    business_types = (
        ('Vendor', 
            ('purchase', 'order', 'payment', 'purchase-credit',), 
            True,
        ),
        ('Client', 
            ('invoice', 'receipt',),
            False,
        ),
        ('Member', 
            ('paycheck',),
            False,
        ),
        ('Creditor', 
            ('credit-card-payment',),
            False,
        ),
    )
    
    for name, exchange_type_slugs, view_all_projects in business_types:
        bt = crm.BusinessType.objects.create(
            name=name,
            can_view_all_projects=view_all_projects
        )
        for slug in exchange_type_slugs:
            bt.exchange_types.add(ExchangeType.objects.get(slug=slug))
    
    print 'Accounts and Exchange Types successfully loaded'
    print
    print '--------------------------------------'
    print 'Install the minibooks SQL like so:'
    print '  createlang plpgsql minibooks'
    print '  ./manage.py dbshell < ledger/sql/minibooks.sql'
    print '--------------------------------------'
    print



# -*- coding: utf-8 -*-
# ----------------------------------------------------------------------------
# $Id$
# ----------------------------------------------------------------------------
#
#    Copyright (C) 2008 Caktus Consulting Group, LLC
#
#    This file is part of minibooks.
#
#    minibooks is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Affero General Public License as
#    published by the Free Software Foundation, either version 3 of 
#    the License, or (at your option) any later version.
#    
#    minibooks is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU Affero General Public License for more details.
#    
#    You should have received a copy of the GNU Affero General Public License
#    along with minibooks.  If not, see <http://www.gnu.org/licenses/>.
#


from django import forms
from django.contrib import admin
from django.contrib.auth.models import User, Group
from django.contrib.auth import admin as admin_app
from crm import models as crm
from minibooks.ledger import models as ledger

books_admin = admin.AdminSite()

class UserAdmin(admin_app.UserAdmin):
    list_display = ('id', 'last_name', 'first_name', 'email', 'last_login', 'is_staff')
    search_fields = ('email', 'first_name', 'last_name',)
    list_filter = ('groups','is_staff')
    ordering = ('last_name',)
books_admin.register(User, UserAdmin)


class GroupAdmin(admin_app.GroupAdmin):
    fields = ('name', 'permissions',)
books_admin.register(Group, GroupAdmin)


class ExchangeTypeAdmin(admin.ModelAdmin):
    list_display = ('label', 'debit', 'credit', 'common_account')
books_admin.register(ledger.ExchangeType, ExchangeTypeAdmin)
    

class BusinessTypeAdmin(admin.ModelAdmin):
    pass
books_admin.register(crm.BusinessType, BusinessTypeAdmin)



# -*- coding: utf-8 -*-
# ----------------------------------------------------------------------------
# $Id: arrayutil.py 310 2008-10-02 17:24:23Z copelco $
# ----------------------------------------------------------------------------
#
#    Copyright (C) 2008 Caktus Consulting Group, LLC
#
#    This file is part of minibooks.
#
#    minibooks is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Affero General Public License as
#    published by the Free Software Foundation, either version 3 of 
#    the License, or (at your option) any later version.
#    
#    minibooks is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU Affero General Public License for more details.
#    
#    You should have received a copy of the GNU Affero General Public License
#    along with minibooks.  If not, see <http://www.gnu.org/licenses/>.
#

import datetime

try:
    from dateutil.relativedelta import relativedelta
except ImportError:
    raise ImportError('minibooks was unable to import the dateutil library. Please confirm it is installed and available on your current Python path.')

from django import template
from django.contrib.humanize.templatetags.humanize import intcomma
from django.template.defaultfilters import floatformat

register = template.Library()

@register.filter(name='currency')
def currency(num):
    return intcomma(floatformat(num, 2))


@register.inclusion_tag('books/ledger/_date_filters.html', takes_context=True)
def ledger_date_filters(context):
    request = context['request']
    from_slug = 'from_date'
    to_slug = 'to_date'
    use_range = True
    
    def construct_url(from_date, to_date):
        url = '%s?%s=%s' % (
            request.path,
            to_slug,
            to_date.strftime('%m/%d/%Y'),
        )
        if use_range:
            url += '&%s=%s' % (
                from_slug,
                from_date.strftime('%m/%d/%Y'),
            )
        return url

    filters = {}
    filters['Past 12 Months'] = []
    single_month = relativedelta(months=1)
    from_date = datetime.date.today().replace(day=1) + relativedelta(months=1)
    for x in range(12):
        to_date = from_date
        from_date = to_date - single_month
        url = construct_url(from_date, to_date - relativedelta(days=1))
        filters['Past 12 Months'].append((from_date.strftime("%b '%y"), url))
    filters['Past 12 Months'].reverse()
    
    start = datetime.date.today().year - 3
    
    filters['Years'] = []
    for year in range(start, start + 3):
        from_date = datetime.datetime(year, 1, 1)
        to_date = from_date + relativedelta(years=1)
        url = construct_url(from_date, to_date - relativedelta(days=1))
        filters['Years'].append((str(from_date.year), url))

    filters['Quaters (Calendar Year)'] = []
    to_date = datetime.date(datetime.date.today().year - 1, 1, 1)
    for x in range(8):
        from_date = to_date
        to_date = from_date + relativedelta(months=3)
        url = construct_url(from_date, to_date - relativedelta(days=1))
        filters['Quaters (Calendar Year)'].append(
            ('Q%s %s' % ((x % 4) + 1, from_date.year), url)
        )

    return {'filters': filters}



#postgres
import psycopg2
import psycopg2.extensions
#regular expressions
import re

from django.shortcuts import render_to_response
from call_log.models import *

#should I be using a manager?
def log(request):
    DSN = "dbname='asterisk' user='asterisk' host='localhost' password='ouFoo0yi'"
    conn = psycopg2.connect(DSN)
    cursor = conn.cursor(cursor_factory=LoggingCursor)
    #cursor = connection.cursor()
    cursor.execute('SELECT * FROM cdr LIMIT 20')
    result_set = []
    rows = cursor.fetchall()
    for row in rows:
        result = {}
        result['acctId'] = row[0]
        #format info: http://www.djangoproject.com/documentation/templates/#now
        result['calldate'] = row[1]
        result['callerid'] = row[2]
        result['src'] = row[3]
        result['dst'] = row[4]
        #dcontext 5
        #channel 6
        #dstchannel 7
        #lastapp 8
        #lastdata 9
        result['duration'] = row[10]
        #bellsec 11
        result['disposition'] = row[12]
        result['amaflags'] = row[13]
        #accountcode 14
        #uniqueid 15
        #userfield 16
        result['callerid_name']   = re.search("\".*\"",result['callerid']).group().strip("\"")
        result['callerid_number'] = re.search("<.*>",result['callerid']).group().strip("<>")
        result_set.append(result)

    context = {
        'calls' : result_set,
        'rows' : rows,
    }

    return render_to_response('call_log/log.html', context)



#!/usr/bin/python

from django.core.management.base import BaseCommand

from minibooks.ledger.demo import setup_demo


class Command(BaseCommand):
    help = "Setup minibooks demo"
    requires_model_validation = False

    def handle(self, **options):
        setup_demo()



# -*- coding: utf-8 -*-
# ----------------------------------------------------------------------------
# $Id$
# ----------------------------------------------------------------------------
#
#    Copyright (C) 2008 Caktus Consulting Group, LLC
#
#    This file is part of minibooks.
#
#    minibooks is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Affero General Public License as
#    published by the Free Software Foundation, either version 3 of 
#    the License, or (at your option) any later version.
#    
#    minibooks is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU Affero General Public License for more details.
#    
#    You should have received a copy of the GNU Affero General Public License
#    along with minibooks.  If not, see <http://www.gnu.org/licenses/>.
#


import os
import shutil
import tempfile
import logging
import subprocess

from django.template.loader import render_to_string
from django.conf import settings


def render_to_pdf_string(path, context):
    """
    Render file (path) to a PDF and return it
    
     * create temporary file
     * write tex to it
     * use Popen to run pdflatex
     * open resulting PDF file and return it
     * cleanup temporary files
    
     -- use file i/o rather than pipes because
        I can't get them to work with batchmode
    """
    
    # create temporary .tex file
    directory_path = tempfile.mkdtemp()
    file_number, file_path = tempfile.mkstemp(
        settings.TEX_EXTENSION,
        '',
        directory_path,
    )
    
    # render template and save to .tex file
    f = os.fdopen(file_number, 'w')
    string = render_to_string(path, context)
    f.write(string)
    f.close()
    
    # run pdflatex on temporary .tex file
    os.chdir(directory_path)
    p = subprocess.Popen(
        '%s %s' % (settings.LATEX_CMD, file_path),
        shell=True,
        stdin=subprocess.PIPE,
        stdout=subprocess.PIPE,
        stderr=subprocess.STDOUT,
    )
    (child_stdin, child_stdout_and_stderr) = (p.stdin, p.stdout)
    pdflatex_output = child_stdout_and_stderr.read()
    if (pdflatex_output.find('Fatal') > 0):
        raise IOError(pdflatex_output)
    os.waitpid(p.pid, 0)
    
    # open resulting PDF and return it
    pdf_file = file_path.replace(
        settings.TEX_EXTENSION,
        settings.PDF_EXTENSION,
    )
    f = open(pdf_file, 'r')
    output = f.read()
    f.close()
    
    shutil.rmtree(directory_path)
    return output



from django.db import models
from django.conf import settings

import logging
import psycopg2
import psycopg2.extensions
import psycopg2.extras
from datetime import datetime, timedelta
#from trac.config import *

import os
import pyparsing
import pprint

from crm.models import Project

class LoggingCursor(psycopg2.extensions.cursor):
    def execute(self, sql, args=None):
        logger = logging.getLogger('sql_debug')
        logger.info(self.mogrify(sql, args))
        
        try:
            psycopg2.extensions.cursor.execute(self, sql, args)
        except Exception, exc:
            logger.error("%s: %s" % (exc.__class__.__name__, exc))
            raise

class TracConnection:
    def __init__(self):
        # make single connection to trac database
        DSN = "dbname='%s' user='%s' password='%s' host='%s'" % ( settings.TRAC_BASE, settings.TRAC_USER, settings.TRAC_PASS, settings.TRAC_HOST )
        psycopg2.extensions.register_type(psycopg2.extensions.UNICODE)
        
        self.connection = psycopg2.connect(DSN)
        self.connection.set_client_encoding('UTF8')

    def get_project_hours_sql(self, projects, min_date, max_date):
        select = ""
        for project in projects:
            if project.trac_environment == None or project.trac_environment == '':
                continue
            project_hours_select = str("""
                SELECT 
                             author,
                       SUM( CASE 
                                                WHEN newvalue = '' OR newvalue is NULL THEN 0 
                                                ELSE newvalue::float 
                                        END ), 
                             value AS billable,
                             '%s' AS project
                       FROM (
                                    SELECT * 
                                    FROM %s.ticket_change, %s.ticket_custom
                                    WHERE %s.ticket_change.ticket = %s.ticket_custom.ticket AND
                                          %s.ticket_custom.name = 'billable'
                             ) AS combined
                             WHERE 
                                      field='hours' AND 
                                      time >= EXTRACT(EPOCH FROM TIMESTAMP '%s') AND 
                                      time <= EXTRACT(EPOCH FROM TIMESTAMP '%s')
                             GROUP BY author, billable\n
            """ % (
                      project.trac_environment, #project column value
                  project.trac_environment, #FROM FROM
                  project.trac_environment, #FROM FROM
                  project.trac_environment, #FROM WHERE
                  project.trac_environment, #FROM WHERE
                  project.trac_environment, #FROM WHERE
                        min_date, 
                        max_date,
                        )
            )
            select += project_hours_select
            if project != list(projects)[-1]:
                select += '\nUNION ALL\n'
        print select
        print "XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX"
        return select


    def getUserTotalHours(self, projects, min_date, max_date, hash_prefix):
        cursor = self.connection.cursor(cursor_factory=LoggingCursor)
        select = self.get_project_hours_sql(projects, min_date=min_date, max_date=max_date)
        select = "SELECT author, SUM(sum), billable FROM ( %s ) AS total GROUP BY author, billable ORDER BY author;" % select
        try:
            cursor.execute(select)
        except Exception, e:
            print 'getTotalHours: select failed: %s' % e
            #print select
            return
        rows = cursor.fetchall()
        checks = {}
        for row in rows:
            if row[2] == '1':
                hash = { hash_prefix + "_billable" : row[1] }
            elif row[2] == '0':
                hash = { hash_prefix + "_not_billable" : row[1] }
            else:
                raise Exception('Unknown billable status', row[2])
            if checks.has_key(row[0]):
                checks[row[0]].update(hash)
            else:
                checks[row[0]] = hash
        return checks


    def getTickets(self, projects, username):
    
        columns = ( 'id', 'cc', 'owner', 'summary', 'changetime' )
        columns_select = ', '.join( column for column in columns )
        
        result = []
        for project in projects:
            if project.trac_environment == None or project.trac_environment == '':
                result.append((project,()))
                continue
            
            try:
                cursor = self.connection.cursor(cursor_factory=LoggingCursor)
                cursor.execute( str("SELECT %s FROM %s.ticket WHERE ( owner ILIKE '%s' OR cc ILIKE '%s' ) AND status!='closed'" % (columns_select.__str__(), project.trac_environment, username.__str__(), username.__str__() ) ) )
            except Exception, e:
                print "getTickets: select failed: %s" % e
                continue
            
            rows = cursor.fetchall()
            
            tickets = []
            for row in rows:
                ticket = {}
                for i in range( len(columns) ):
                    if 'time' in columns[i]:
                        ticket.update( { columns[i] : datetime.fromtimestamp(row[i]) } )
                    else:
                        ticket.update( { columns[i] : row[i] } )
                if ticket['id']:
                    ticket.update( { 'url' : '%sticket/%s' % ( project.trac_url(), ticket['id'] ) } )
                tickets.append(ticket)
            result.append( (project, tickets ) )
    
        return result




# -*- coding: utf-8 -*-
# ----------------------------------------------------------------------------
# $Id$
# ----------------------------------------------------------------------------
#
#    Copyright (C) 2008 Caktus Consulting Group, LLC
#
#    This file is part of minibooks.
#
#    minibooks is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Affero General Public License as
#    published by the Free Software Foundation, either version 3 of 
#    the License, or (at your option) any later version.
#    
#    minibooks is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU Affero General Public License for more details.
#    
#    You should have received a copy of the GNU Affero General Public License
#    along with minibooks.  If not, see <http://www.gnu.org/licenses/>.
#


import datetime

try:
    from dateutil.relativedelta import relativedelta
except ImportError:
    raise ImportError('minibooks was unable to import the dateutil library. Please confirm it is installed and available on your current Python path.')

from urllib import urlencode

from django.contrib.admin.views.decorators import staff_member_required
from django.contrib.auth.decorators import login_required, permission_required
from django.core.urlresolvers import reverse
from django.core.mail import EmailMessage
from django.db import transaction
from django.http import HttpResponseRedirect, HttpResponse, QueryDict
from django.shortcuts import render_to_response, get_object_or_404
from django.template import RequestContext
from django.template.loader import render_to_string
from django.utils import simplejson as json
from django.db.models import Q
from django.contrib.auth.models import User
from django.conf import settings

from caktus.django.decorators import render_with
from caktus.django.db.util import query_to_table
from caktus.django.db.util import search_path as model_search_path
from caktus.django.forms import AutoCompleterForm

from crm import models as crm
from crm import forms as crm_forms

from minibooks.ledger import models as ledger
from minibooks.ledger import forms as ledger_forms
from minibooks.ledger.report import generate_exchange_pdf, generate_project_report_pdf
from minibooks.ledger.communication import send_project_report_email, send_exchange_email
from minibooks.ledger.report import generate_exchange_pdf, generate_project_report_pdf
from minibooks.ledger.models import Exchange, Account


@login_required
@render_with('books/ledger/index.html')
def dashboard(request):
    # there are no permissions on this view, so all DB access
    # must filter by request.user
    exchanges = ledger.Exchange.objects.filter(
        business__contacts=request.user
    ).select_related('type', 'business')[:10]
    
    # soonest first
    upcoming_interactions = request.user.interactions.select_related(
        'cdr',
        'contacts',
        'project',
        'project__business',
    ).filter(completed=False).order_by('date')
    
    # most recent first
    recent_interactions = request.user.interactions.select_related(
        'cdr',
        'contacts',
        'project',
        'project__business',
    ).filter(completed=True)[:6]
    
    projects = request.user.contact_projects.order_by(
        'name',
    ).select_related('business')
    
    context = {
        'recent_exchanges': exchanges,
        'recent_interactions': recent_interactions,
        'upcoming_interactions': upcoming_interactions,
        'projects': projects,
    }
    
    return context


@permission_required('ledger.view_exchange')
@render_with('books/ledger/exchange/list.html')
def list_exchanges(request, exchange_type_slug=None):
    if exchange_type_slug:
        exchange_type = get_object_or_404(
            ledger.ExchangeType, 
            slug=exchange_type_slug,
        )
    else:
        exchange_type = None
    
    if exchange_type:
        exchanges = ledger.Exchange.objects.filter(type=exchange_type)
    else:
        exchanges = ledger.Exchange.objects
    
    searching = False
    form = ledger_forms.SearchForm(request.GET)
    if form.is_valid() and form.cleaned_data['search']:
        searching = True
        search = form.cleaned_data['search']
        exchanges = exchanges.filter(
            Q(business__name__icontains=search) |
            Q(memo__icontains=search) |
            Q(transactions__memo__icontains=search) |
            Q(transactions__project__name__icontains=search)
        ).distinct()
    
    if not searching and not exchange_type:
        sql = []
        exchanges_types = ledger.ExchangeType.objects.all()
        for t in exchanges_types:
            sql.append(
                "(SELECT id FROM ledger_exchange WHERE type_id=%s ORDER BY date DESC LIMIT 5)" % t.id
            )
        exchanges = exchanges.extra(
            where=['ledger_exchange.id IN (%s)' % " UNION ".join(sql)]
        )
    
    exchanges = exchanges.select_related(
        'business',
        'type',
    ).order_by('type', '-date', '-id',)
    
    context = {
        'form': form,
        'searching': searching,
        'exchange_type': exchange_type,
        'exchanges': exchanges,
    }
    return context


@permission_required('ledger.view_exchange')
def view_exchange_as_pdf(request, exchange_id):
    exchange = get_object_or_404(ledger.Exchange, id=exchange_id)
    pdf = generate_exchange_pdf(exchange)
    
    return HttpResponse(pdf, content_type='application/pdf')


@permission_required('ledger.email_exchange')
@transaction.commit_on_success
@render_with('books/ledger/exchange/email/exchange.html')
def new_exchange_email(request, exchange_id):
    try:
        exchange = Exchange.objects.select_related().get(pk=exchange_id)
    except Exchange.DoesNotExist:
        raise Http404
    
    project = exchange.get_project()
    
    if request.method == 'POST':
        form = ledger_forms.EmailForm(
            request.POST, 
            business=exchange.business, 
            project=project,
        )
        if form.is_valid():
            memo = form.cleaned_data['memo']
            recipient = form.cleaned_data['to']
            context = {
                "exchange": exchange,
                "memo": memo,
                "recipient": recipient,
            }
            recipients = [recipient.email,]
            attachment = generate_exchange_pdf(exchange)
            if project:
                subject = project.name + ' ' + exchange.type.label
            elif exchange.business:
                subject = exchange.business.name + ' ' + exchange.type.label
            else:
                subject = exchange.type.label
            subject += ' %s' % exchange.date
            
            if not send_exchange_email(
                exchange_type=exchange.type,
                recipients=recipients,
                subject=subject,
                context=context,
                attachment=attachment,
            ):
                request.notifications.add('Failed to send email')
                return HttpResponseRedirect(
                    reverse(
                        'view_business',
                        kwargs={'business_id': exchange.business.id},
                    )
                )
            
            # create and save interaction
            message = 'Successfully sent %s to %s (%s)' % (
                exchange.type.label,
                recipient.get_full_name(),
                recipient.email,
            )
            interaction = crm.Interaction.objects.create(
                date=datetime.datetime.now(),
                type='business',
                completed=True,
                memo='%s: %s' % (message, memo),
            )
            exchange.delivered = True
            exchange.save()
            
            # add projects contacts to contacts list
            if project:
                for contact in project.contacts.all():
                    interaction.contacts.add(contact)
            elif exchange.business:
                for contact in exchange.business.contacts.all():
                    interaction.contacts.add(contact)
            
            request.notifications.add(message)
            
            return HttpResponseRedirect(
                reverse(
                    'view_business',
                    kwargs={'business_id': exchange.business.id},
                )
            )
    else:
        form = ledger_forms.EmailForm(
            business=exchange.business, 
            project=project,
        )
    
    context = {
        'exchange': exchange,
        'form': form,
    }
    return context


@permission_required('crm.email_project_report')
@transaction.commit_on_success
@render_with('books/ledger/exchange/email/project_report.html')
def new_project_report_email(request, project):
    if request.method == 'POST':
        form = ledger_forms.EmailForm(request.POST, project=project)
        if form.is_valid():
            memo = form.cleaned_data['memo']
            recipient = form.cleaned_data['to']
            
            context = {
                "project": project,
                "memo": memo,
                "recipient": recipient,
            }
            recipients = [recipient.email,]
            attachment = generate_project_report_pdf(project)
            subject = "%s project report" % project.name
            
            if not send_project_report_email(
                recipients=recipients,
                context=context,
                attachment=attachment
            ):
                request.notifications.add('Failed to send email')
                return HttpResponseRedirect(
                    reverse(
                        'view_project',
                        kwargs={
                            'business_id': project.business.id,
                            'project_id': project.id,
                        },
                    )
                )
            
            message = 'Successfully sent Project Report to %s (%s)' % (
                recipient.get_full_name(),
                recipient.email,
            )
            interaction = crm.Interaction.objects.create(
                date=datetime.datetime.now(),
                type='business',
                completed=True,
                project=project,
                memo='%s: %s' % (message, memo),
            )
            # add projects contacts to contacts list
            for contact in interaction.project.contacts.all():
                interaction.contacts.add(contact)
            request.notifications.add(message)
            
            return HttpResponseRedirect(
                reverse(
                    'view_project',
                    kwargs={
                        'business_id': project.business.id,
                        'project_id': project.id,
                    },
                )
            )
    else:
        form = ledger_forms.EmailForm(project=project)
    
    context = {
        'project': project,
        'form': form,
    }
    return context


@permission_required('ledger.add_exchange')
@permission_required('ledger.change_exchange')
@transaction.commit_on_success
@render_with('books/ledger/exchange/create.html')
def create_edit_exchange(
    request,
    exchange_type_slug,
    business=None,
    project=None,
    exchange_id=None,
  ):
    exchange_type = get_object_or_404(
        ledger.ExchangeType, 
        slug=exchange_type_slug
    )
    exchange = None
    repeat_period = None
    
    if exchange_id:
        exchange = get_object_or_404(
            ledger.Exchange, 
            pk=exchange_id, 
            type=exchange_type,
        )
        repeat_period = exchange.repeat_period
    
    repeat_period_form = ledger_forms.RepeatPeriodForm(
        request,
        instance=repeat_period,
        prefix='repeat',
    )
    exchange_form = ledger_forms.ExchangeForm(
        request,
        instance=exchange,
        exchange_type=exchange_type,
    )
    
    if request.POST:
        transaction_formset = ledger_forms.TransactionFormSet(
            request.POST,
            # workaround for http://code.djangoproject.com/ticket/9462
            instance=exchange or ledger.Exchange(), 
            exchange_type=exchange_type,
        )
        
        if (request.POST
            and exchange_form.is_valid()
            and transaction_formset.is_valid()
            and repeat_period_form.is_valid()
          ):
            exchange = exchange_form.save(commit=False)
            exchange.repeat_period = repeat_period_form.save()
            exchange.save()
            
            def save_transaction(transaction):
                transaction.exchange = exchange
                debit_or_credit = exchange_type.debit_or_credit()
                setattr(
                    transaction,
                    debit_or_credit,
                    exchange_form.cleaned_data[debit_or_credit],
                )
                if project:
                    transaction.project = project
                else:
                    transaction.project = \
                        exchange_form.cleaned_data['project']
                transaction.save()
            
            transactions = transaction_formset.save(commit=False)
            for transaction in transactions:
                save_transaction(transaction)
                
            ids = [t.id for t in transactions]
            for transaction in exchange.transactions.exclude(id__in=ids):
                save_transaction(transaction)
                
            request.user.message_set.create(
                message='%s successfully saved.' % exchange_type,
            )
            return HttpResponseRedirect(
                reverse(
                    'list_exchanges_by_type',
                    kwargs={'exchange_type_slug': exchange.type.slug},
                )
            )
    else:
        transaction_formset = ledger_forms.TransactionFormSet(
            instance=exchange, 
            exchange_type=exchange_type,
        )

    context = {
        'exchange': exchange,
        'exchange_type': exchange_type,
        'exchange_form': exchange_form,
        'transaction_formset': transaction_formset,
        'repeat_period_form': repeat_period_form,
    }
    return context


@permission_required('ledger.change_exchange')
@transaction.commit_on_success
@render_with('books/ledger/exchange/edit_exchange_type.html')
def edit_exchange_type(request, exchange_type_slug, exchange_id):
    exchange_type = get_object_or_404(
      ledger.ExchangeType, 
      slug=exchange_type_slug,
    )
    exchange = get_object_or_404(
      ledger.Exchange, 
      pk=exchange_id, 
      type=exchange_type,
    )
    exchange_form = ledger_forms.ExchangeTypeForm(
        request,
        instance=exchange,
    )
    if request.POST and exchange_form.is_valid():
        exchange = exchange_form.save()
        request.notifications.add(
            'Successfully change exchange type from %s to %s' % (
                exchange_type,
                exchange.type,
            )
        )
        return HttpResponseRedirect(
            reverse('edit_exchange', kwargs={
                'exchange_id': exchange.id,
                'exchange_type_slug': exchange.type.slug,
            })
        )
    context = {
        'exchange': exchange,
        'exchange_form': exchange_form,
    }
    return context

@permission_required('ledger.delete_exchange')
@render_with('books/ledger/exchange/remove.html')
@transaction.commit_on_success
def remove_exchange(request, exchange_id):
    exchange = get_object_or_404(ledger.Exchange, pk=exchange_id)
    
    if request.POST and not exchange.reconciled:
        exchange.delete()
        return HttpResponseRedirect(reverse('list_exchanges'))
    
    context = {
        'reconciled': exchange.reconciled,
        'exchange': exchange,
    }
    return context


@permission_required('ledger.add_exchange')
@permission_required('ledger.change_exchange')
@transaction.commit_on_success
def duplicate_exchange(request, exchange_id):
    exchange = get_object_or_404(ledger.Exchange, pk=exchange_id)
    transactions = exchange.transactions.all()
    
    exchange.id = None
    exchange.memo = 'Copy of %s' % (exchange.memo)
    exchange.delivered = False
    exchange.save(force_insert=True)
    for transaction in transactions:
        transaction.id = None
        transaction.memo = 'Copy of %s' % (transaction.memo)
        transaction.exchange = exchange
        transaction.debit_reconciled = False
        transaction.credit_reconciled = False
        transaction.save(force_insert=True)
    
    if request.GET['next']:
        return HttpResponseRedirect(request.GET['next'])
    else:
        return HttpResponseRedirect(reverse('list_exchanges'))


### not a view
def date_filters(
    request,
    from_slug='from_date',
    to_slug='to_date',
    use_range=True,
  ):
    def construct_url(from_date, to_date):
        url = '%s?%s=%s' % (
            request.path,
            to_slug,
            to_date.strftime('%m/%d/%Y'),
        )
        if use_range:
            url += '&%s=%s' % (
                from_slug,
                from_date.strftime('%m/%d/%Y'),
            )
        return url
        
    filters = {}
    filters['Past 12 Months'] = []
    single_month = relativedelta(months=1)
    from_date = datetime.date.today().replace(day=1) + relativedelta(months=1)
    for x in range(12):
        to_date = from_date
        from_date = to_date - single_month
        url = construct_url(from_date, to_date - relativedelta(days=1))
        filters['Past 12 Months'].append((from_date.strftime("%b '%y"), url))
    filters['Past 12 Months'].reverse()
    
    filters['Years'] = []
    for year in ledger.Transaction.objects.dates('date', 'year'):
        from_date = year
        to_date = year + relativedelta(years=1)
        url = construct_url(from_date, to_date - relativedelta(days=1))
        filters['Years'].append((str(from_date.year), url))
    
    filters['Quaters (Calendar Year)'] = []
    to_date = datetime.date(datetime.date.today().year - 1, 1, 1)
    for x in range(8):
        from_date = to_date
        to_date = from_date + relativedelta(months=3)
        url = construct_url(from_date, to_date - relativedelta(days=1))
        filters['Quaters (Calendar Year)'].append(
            ('Q%s %s' % ((x % 4) + 1, from_date.year), url)
        )
    
    return filters


@permission_required('ledger.view_account')
@render_with('books/ledger/account/ledger.html')
def account_ledger(request, account_id=None):
    if account_id:
        accounts = [ get_object_or_404(ledger.Account, pk=account_id) ]
    else:
        if request.GET and 'accounts' in request.GET:
            accounts = ledger.Account.objects.filter(
                id__in=request.GET.getlist('accounts')
            )
        else:
            accounts = ledger.Account.objects.all()
    
    date_form = ledger_forms.DateForm(request, use_get=True)
    if request.GET and date_form.is_valid():
        from_date, to_date = date_form.save()
    else:
        from_date = None
        to_date = None
    
    context = {
        'accounts': accounts,
        'reconcile_enabled': bool(account_id),
        'filters': date_filters(request),
        'to_date': to_date,
        'from_date': from_date,
        'date_form': date_form,
    }
    return context


@permission_required('ledger.view_account')
@render_with('books/ledger/account/business_balance.html')
def account_business_balances(request, account_id):
    account = get_object_or_404(ledger.Account, pk=account_id)
    
    date_form = ledger_forms.SingleDateForm(request, use_get=True)
    if request.GET and date_form.is_valid():
        date = date_form.save()
    else:
        date = None
    
    def sum_by_business(debit=None, credit=None, to_date=None):
        args = []
        where = []
        
        if debit:
            where.append("t.debit_id = %s")
            args.append(debit.id)
        if credit:
            where.append("t.credit_id = %s")
            args.append(credit.id)
        if to_date:
            where.append("t.date <= %s")
            args.append(to_date)
            
        sql = """
    SELECT
        c.name,
        e.business_id,
        COALESCE(SUM(ROUND(amount * quantity, 2)), 0.0) 
    FROM ledger_transaction t
    JOIN ledger_exchange e ON (t.exchange_id = e.id)
    JOIN crm_contact c ON (c.id = e.business_id AND c.type = 'business')
    """
        if where:
            sql += " WHERE %s" % ' AND '.join(where)
            
        sql += " GROUP BY e.business_id, c.name ORDER BY c.name"
        
        from django.db import connection
        cursor = connection.cursor()
        cursor.execute(sql, args)
        result = {}
        for name, id, total in cursor.fetchall():
            result[id] = (name, total)
        return result
    
    credit_totals = sum_by_business(credit=account, to_date=date)
    debit_totals = sum_by_business(debit=account, to_date=date)
    
    balances = []
    for id in set(credit_totals.keys() + debit_totals.keys()):
        credit_name, credit_total = credit_totals.get(id, ('', 0))
        debit_name, debit_total = debit_totals.get(id, ('', 0))
        if account.type in ('asset', 'expense', 'equity',):
            balance = debit_total - credit_total
        else:
            balance = credit_total - debit_total
        balances.append({
            'id': id,
            'name': credit_name or debit_name,
            'debit_total': debit_total,
            'credit_total': credit_total,
            'balance': balance,
        })
    
    context = {
        'account': account,
        'balances': balances,
        'filters': date_filters(request, to_slug='date', use_range=False),
        'to_date': date,
        'date_form': date_form,
    }
    return context


@permission_required('ledger.add_account')
@permission_required('ledger.change_account')
@transaction.commit_on_success
@render_with('books/ledger/account/create_edit.html')
def create_edit_account(request, account_id=None):
    if account_id:
        account = ledger.Account.objects.get(pk=account_id)
    else:
        account = None
    
    if request.POST:
        account_form = ledger_forms.AccountForm(request.POST, instance=account)
        if account_form.is_valid():
            account_form.save()
            return HttpResponseRedirect(reverse('list_accounts'))
        
    else:
        account_form = ledger_forms.AccountForm(instance=account)
    
    context = {
        'account': account,
        'account_form': account_form,
    }
    return context

@permission_required('ledger.delete_account')
@transaction.commit_on_success
@render_with('books/ledger/account/remove.html')
def remove_account(request, account_id):
    account = get_object_or_404(ledger.Account, pk=account_id)
    transaction_count = account.get_transactions().count()
    
    if transaction_count == 0 and request.POST:
        account.delete()
        return HttpResponseRedirect(reverse('list_accounts'))
    
    context = {
        'account': account,
        'transaction_count': transaction_count,
    }
    return context


@permission_required('ledger.view_account')
@render_with('books/ledger/account/list.html')
def list_accounts(request):
    context = {
        'accounts': ledger.Account.objects.all(),
    }
    return context


@permission_required('crm.view_project')
def project_report(request, project):
    ar = ledger.Account.objects.get(number=1200)
    credit_total = ar.credit_total_for_project(project.id)
    debit_total = ar.debit_total_for_project(project.id)
    credits = ar.credits.filter(project=project.id)
    debits = ar.debits.filter(project=project.id)
    client = project.business
    
    if not credit_total: credit_total = 0
    if not debit_total:   debit_total = 0
    outstanding_balance = debit_total - credit_total
    
    return render_to_response('books/report/report.html', locals())


@permission_required('crm.view_project')
def show_project_report(request, project):
    pdf = generate_project_report_pdf(project)
    
    return HttpResponse(pdf, content_type='application/pdf')


@permission_required('ledger.change_transaction')
def reconcile_transaction(request):
    value = False
    reconciled_balance = '0.0'
    errors = []
    
    if request.POST:
        form = ledger_forms.ReconcileTransactionForm(request.POST)
        if form.is_valid():
            try:
                transaction = form.cleaned_data['transaction']
                account = form.cleaned_data['account']
                value = not getattr(transaction, '%s_reconciled' % account)
                setattr(transaction, '%s_reconciled' % account, value)
                transaction.save()
                account_id = getattr(transaction, '%s_id' % account)
                account = ledger.Account.objects.get(pk=account_id)
                to_date = form.cleaned_data['to_date']
                reconciled_balance = \
                  '%.2f' % account.reconciled_balance(to_date=to_date)
            except Exception, e:
                errors.append('%s: %s' % (type(e).__name__, e.message))
        else:
            for k, v in form.errors.iteritems():
                errors.append('%s: %s' % (k, v))
    else:
        errors.append('Must use form POST to reconcile a transaction')
    
    if errors:
        errors = '</li><li>'.join(errors)
        errors = '<ul class="errorlist"><li>%s</li></ul>' % errors
    else:
        errors = ''
    
    response = {
        'reconciled': value,
        'reconciled_balance': reconciled_balance,
        'errors': errors,
    }
    return HttpResponse(json.dumps(response), mimetype="text/json")


@permission_required('ledger.transfer_funds')
def list_transfer_debit_accounts(request):
    results = []
    
    if 'credit_id' in request.GET:
        credit = ledger.Account.objects.get(id=request.GET['credit_id'])
        accounts = ledger.Account.objects.filter(type=credit.type)
        accounts = accounts.exclude(id=credit.id)
        for account in accounts:
            results.append({
                'value': account.id,
                'label': unicode(account),
            })
    
    return HttpResponse(json.dumps(results), mimetype="text/json")


@permission_required('ledger.transfer_funds')
@transaction.commit_on_success
@render_with('books/ledger/account/transfer.html')
def transfer_funds(request, transaction_id=None):
    if transaction_id:
        transaction = get_object_or_404(ledger.Transaction, id=transaction_id)
    else:
        transaction = None
    
    if request.POST:
        form = ledger_forms.TransferForm(request.POST, instance=transaction)
        if form.is_valid():
            form.save()
            return HttpResponseRedirect(reverse('list_accounts'))
    else:
        form = ledger_forms.TransferForm(instance=transaction)
    
    context = {
        'form': form,
    }
    return context


@permission_required('ledger.add_transaction')
@permission_required('ledger.change_transaction')
@transaction.commit_on_success
@render_with('books/ledger/account/general_entry.html')
def general_entry(request, transaction_id=None):
    transaction = None
    if transaction_id:
        transaction = get_object_or_404(
            ledger.Transaction,
            id=transaction_id,
            project__isnull=True,
            exchange__isnull=True,
        )
    form = ledger_forms.GeneralEntryForm(request, instance=transaction)
    if request.POST and form.is_valid():
        form.save()
        return HttpResponseRedirect(reverse('list_accounts'))
    context = {
        'form': form,
        'transaction': transaction,
    }
    return context


@permission_required('ledger.view_account_reports')
@render_with('books/ledger/reports/profit_loss.html')
def profit_loss(request):
    date_form = ledger_forms.DateForm(request, use_get=True)
    if request.GET and date_form.is_valid():
        from_date, to_date = date_form.save()
    else:
        from_date = None
        to_date = None
    
    total_profits = 0
    total_losses = 0
    profit_accounts = ledger.Account.objects.filter(
        type='income',
    ).order_by('name')
    loss_accounts = ledger.Account.objects.filter(
        type='expense',
    ).order_by('name')
    
    for account in profit_accounts:
        account.date_range_total = account.total_for_date_range(
            from_date=from_date,
            to_date=to_date,
        )
        total_profits += account.date_range_total
    
    for account in loss_accounts:
        account.date_range_total = account.total_for_date_range(
            from_date=from_date,
            to_date=to_date,
        )
        total_losses += account.date_range_total
    
    # for links to account pages
    date_query = {}
    if to_date:
        date_query['to_date'] = to_date.strftime('%m/%d/%Y')
    if from_date:
        date_query['from_date'] = from_date.strftime('%m/%d/%Y')
    date_query_string = urlencode(date_query)
    
    context = {
        'date_query_string': date_query_string,
        'date_form': date_form,
        'to_date': to_date,
        'filters': date_filters(request),
        'from_date': from_date,
        'profit_accounts': profit_accounts,
        'loss_accounts': loss_accounts,
        'total_profits': total_profits,
        'total_losses': total_losses,
        'net_income': total_profits - total_losses,
    }
    return context


@permission_required('ledger.view_account_reports')
@render_with('books/ledger/reports/balance_sheet.html')
def balance_sheet(request):
    date_form = ledger_forms.SingleDateForm(request, use_get=True)
    if request.GET and date_form.is_valid():
        date = date_form.save()
    else:
        date = None
    
    total_assets = 0
    total_liabilities = 0
    asset_accounts = ledger.Account.objects.filter(
        type='asset',
    ).order_by('name')
    liability_accounts = ledger.Account.objects.filter(
        type='liability',
    ).order_by('name')
    
    for account in asset_accounts:
        account.date_range_total = account.total_for_date_range(
            to_date=date,
        )
        total_assets += account.date_range_total
    
    for account in liability_accounts:
        account.date_range_total = account.total_for_date_range(
            to_date=date,
        )
        total_liabilities += account.date_range_total
    
    # for links to account pages
    date_query = {}
    if date:
        date_query['to_date'] = date.strftime('%m/%d/%Y')
    date_query_string = urlencode(date_query)
    
    context = {
        'date_query_string': date_query_string,
        'date_form': date_form,
        'filters': date_filters(request, to_slug='date', use_range=False),
        'date': date,
        'asset_accounts': asset_accounts,
        'liability_accounts': liability_accounts,
        'total_assets': total_assets,
        'total_liabilities': total_liabilities,
        'owners_equity': total_assets - total_liabilities,
        'liabilities_and_owners_equity': total_assets,
    }
    return context


@permission_required('ledger.view_account_reports')
@render_with('books/ledger/reports/list.html')
def list_reports(request):
    today = datetime.date.today()
    dates = (
        ('This Month',
            {
                'from_date': today.replace(day=1),
                'to_date': today,
            }
        ),
        ('Last Month',
            {
                'from_date': today.replace(day=1) - relativedelta(months=1),
                'to_date': today.replace(day=1),
            }
        ),
        ('This Year',
            {
                'from_date': today.replace(day=1, month=1),
                'to_date': today,
            }
        ),
        ('Last Year',
            {
                'from_date': \
                    today.replace(day=1, month=1) - relativedelta(years=1),
                'to_date': today.replace(day=1, month=1),
            }
        ),
    )
    headers = []
    revenue = []
    expenses = []
    totals = []
    for label, date_range in dates:
        headers.append(label)
        income = ledger.Account.sum_by_type('income', **date_range)
        revenue.append(income)
        expense = ledger.Account.sum_by_type('expense', **date_range)
        expenses.append(expense)
        totals.append(income - expense)

    context = {
        'summary': {
            'headers': headers,
            'revenue': revenue,
            'expenses': expenses,
            'totals': totals,
        }
    }
    return context


@render_with('books/ledger/cron.html')
def cron(request):
    from minibooks.ledger.cron import create_recurring_exchanges
    messages = create_recurring_exchanges()
    
    if messages:
        email = EmailMessage(
            'minibooks cron messages', 
            '\n'.join(messages), 
            settings.DEFAULT_FROM_EMAIL, 
            ['%s <%s>' % (name, email) for name, email in settings.ADMINS],
        )
        email.send()
    else:
        messages.append('nothing done')
    
    if request.user.is_staff:
        context = {
            'messages': messages,
        }
    else:
        context = {}
    
    return context



# -*- coding: utf-8 -*-
# ----------------------------------------------------------------------------
# $Id$
# ----------------------------------------------------------------------------
#
#    Copyright (C) 2008 Caktus Consulting Group, LLC
#
#    This file is part of minibooks.
#
#    minibooks is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Affero General Public License as
#    published by the Free Software Foundation, either version 3 of 
#    the License, or (at your option) any later version.
#    
#    minibooks is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU Affero General Public License for more details.
#    
#    You should have received a copy of the GNU Affero General Public License
#    along with minibooks.  If not, see <http://www.gnu.org/licenses/>.
#


from django.conf.urls.defaults import *
from django.conf import settings
from django.contrib import admin
from django.views.generic.simple import redirect_to

from caktus.django.views import login, logout, reset_password, change_password

from ledger.admin import books_admin
from crm.xmlrpc import rpc_handler

handler404 = 'caktus.django.views.page_not_found'
handler500 = 'caktus.django.views.server_error'

admin.autodiscover()

urlpatterns = patterns('',
    url(r'^%sbooks-admin/(.*)' % settings.URL_PREFIX, books_admin.root),
    url(r'^%sbooks-admin/$' % settings.URL_PREFIX, books_admin.root, name='books_admin'),
    
    url(r'^%sxml-rpc/' % settings.URL_PREFIX, rpc_handler, name='xml_rpc'),
    
    (r'^%sadmin/(.*)' % settings.URL_PREFIX, admin.site.root),
    
    url(
        r'^%saccount/login/$' % settings.URL_PREFIX,
        login,
        {'redirect_to': settings.LOGIN_REDIRECT_URL},
        name='auth_login',
    ),
    url(
        r'^%saccount/logout/$' % settings.URL_PREFIX,
        logout,
        name='auth_logout',
    ),
    url(
        r'^%saccount/password/change/$' % settings.URL_PREFIX,
        change_password,
        name='change_password',
    ),
    url(
        r'^%saccount/password/reset/$' % settings.URL_PREFIX,
        reset_password,
        {'redirect_to': settings.LOGIN_URL},
        name='reset_password',
    ),
    (r'^%sajax/' % settings.URL_PREFIX, include('ajax_select.urls')),
    
    (r'^%sledger/' % settings.URL_PREFIX, include('minibooks.ledger.urls')),
    (r'^%scrm/' % settings.URL_PREFIX, include('crm.urls')),
    (r'^%scontactinfo/' % settings.URL_PREFIX, include('contactinfo.urls')),
    (r'^%stimepiece/' % settings.URL_PREFIX, include('timepiece.urls')),
    
    # url(
    #     r'^%s(?:home/)?$' % settings.URL_PREFIX,
    #     redirect_to,
    #     kwargs={'url': '/%scrm/dashboard/' % settings.URL_PREFIX}
    # ),
    
    # url(r'^%s' % settings.URL_PREFIX, include('caktus.django.cms.urls')),
    
)




import unittest

from django.db import transaction
from django.test.client import Client

from minibooks.ledger.management import clean_minibooks_db


class TestCase(unittest.TestCase):
    def setUp(self):
        self.c = Client()
        clean_minibooks_db()
    
    def tearDown(self):
        self.assertUserLogout()
        self.c = None
        transaction.rollback()
    
    def assertUserLogout(self):
        # log out only if logged in
        if self.c.cookies:
            self.c.logout()
        self.user = None
    
    def assertUserLogin(self, username, password):
        login_successful = self.c.login(username=username, password=password)
        self.assertEqual(login_successful, True)
        try:
            self.user = User.objects.get(email=username)
        except User.DoesNotExist:
            try:
                self.user = User.objects.get(username=username)
            except User.DoesNotExist:
                pass
    
    def assertRedirects(
        self,
        response,
        expected_path,
        status_code=302,
        target_status_code=200,
      ):
        self.assertEqual(response.status_code, status_code)
        expected_url = "http://testserver%s" % expected_path
        actual_url = response['Location'][0:len(expected_url)]
        self.assertEqual(actual_url, expected_url, str(response))

        if target_status_code:
            response = self.c.get(expected_path)
            self.assertEqual(
                response.status_code,
                target_status_code,
                response,
            )



from sqlalchemy import create_engine
from sqlalchemy import Table, Column, Float, String, Date, MetaData 
from sqlalchemy.orm import mapper, sessionmaker
from sqlalchemy.sql import select
from django.conf import settings

from PIL import Image
from matplotlib.backends.backend_agg import FigureCanvasAgg as FigureCanvas
from matplotlib.figure import Figure
from pylab import date2num
from StringIO import StringIO

engine = create_engine('postgres://%s:%s@%s/%s' % (settings.TRAC_USER, settings.TRAC_PASS, settings.TRAC_HOST, settings.TRAC_BASE), echo=True)
metadata = MetaData()

author_hours_daily_summary = Table('author_hours_daily_summary', metadata,
        Column('date', Date, primary_key=True),
        Column('author', String(40)),
        Column('hours', Float )
)

class DayAuthorHoursSummary(object):
    def __init__(self, author, date, hours):
        self.author = author
        self.date = date
        self.hours = hours
        
    def __repr__(self):
        return "<Day('%s', '%s', '%f')>" % (self.author, self.date, self.hours)

mapper(DayAuthorHoursSummary, author_hours_daily_summary)

Session = sessionmaker(bind=engine, autoflush=True, transactional=True)
session = Session()

def developers():
    """Return a list of authors for caktus projects"""
    connection = engine.connect()
    s = select([author_hours_daily_summary.c.author]).distinct()
    result = connection.execute(s)
    authors = []
    for row in result:
        if row[0] != 'root':
            authors.append(row[0])
    result.close()
    return authors



# -*- coding: utf-8 -*-
# ----------------------------------------------------------------------------
# $Id: $
# ----------------------------------------------------------------------------
#
#    Copyright (C) 2008 Caktus Consulting Group, LLC
#
#    This file is part of minibooks.
#
#    minibooks is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Affero General Public License as
#    published by the Free Software Foundation, either version 3 of 
#    the License, or (at your option) any later version.
#    
#    minibooks is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU Affero General Public License for more details.
#    
#    You should have received a copy of the GNU Affero General Public License
#    along with minibooks.  If not, see <http://www.gnu.org/licenses/>.
#

import datetime

from django.db import transaction
from django.core.urlresolvers import reverse
from django.conf import settings

from minibooks.ledger import models as ledger

@transaction.commit_on_success
def create_recurring_exchanges():
    messages = []
    for repeat_period in ledger.RepeatPeriod.objects.filter(active=True):
        exchange = repeat_period.exchanges.order_by('-date').select_related()[0]
        transactions = exchange.transactions.all()
        while exchange.date + repeat_period.delta() <= datetime.date.today():
            exchange.id = None
            exchange.date += repeat_period.delta()
            if exchange.date_due:
                exchange.date_due += repeat_period.delta()
            exchange.delivered = False
            exchange.save(force_insert=True)
            url_kwargs = {
                'exchange_id': exchange.id,
                'exchange_type_slug': exchange.type.slug,
            }
            
            messages.append('%s%s' % (
                settings.APP_URL_BASE,
                reverse('edit_exchange', kwargs=url_kwargs),
            ))
            for transaction in transactions:
                transaction.id = None
                transaction.exchange = exchange
                transaction.date = exchange.date
                transaction.debit_reconciled = False
                transaction.credit_reconciled = False
                transaction.save(force_insert=True)
    
    if messages:
        messages.insert(0, 'created the following exchanges:')
    return messages



import datetime
import random

from django.core.urlresolvers import reverse
from django.test import TestCase
from django.contrib.auth.models import User
from django.template.defaultfilters import slugify
from django.core import mail

from crm import models as crm
from timepiece.tests import TimepieceDataTestCase

from minibooks.ledger import models as ledger
from minibooks.ledger.test import TestCase as BaseTestCase


class LedgerDataTestCase(TimepieceDataTestCase):
    def create_account(self, data={}):
        defaults = {
            'name': self.random_string(20, extra_chars=' '),
            'number': random.randint(1000, 9999),
        }
        defaults.update(data)
        return ledger.Account.objects.create(**defaults)
    
    def create_exchange_type(self, data={}):
        label = self.random_string(20, extra_chars=' ')
        defaults = {
            'label': label,
            'slug': slugify(label),
        }
        defaults.update(data)
        return ledger.ExchangeType.objects.create(**defaults)
    
    def create_exchange(self, data={}):
        defaults = {
            'business': self.create_business(),
            'type': self.create_exchange_type(),
            'date': datetime.datetime.now(),
        }
        defaults.update(data)
        return ledger.Exchange.objects.create(**defaults)

    def create_transaction(self, data={}):
        defaults = {
            'memo': self.random_string(100, extra_chars=' '),
            'amount': '%d.%02d' % (random.randint(1,100000), random.random()),
            'quantity': random.randint(1,10000),
            'date': datetime.datetime.now(),
        }
        if 'debit' not in data:
            data['debit'] = self.create_account()
        if 'credit' not in data:
            data['credit'] = self.create_account()
        defaults.update(data)
        return ledger.Transaction.objects.create(**defaults)


class ExchangeEmails(LedgerDataTestCase):
    def setUp(self):
        self.superuser = User.objects.create_user('superuser', 
                                                  'test@example.com', 'test')
        self.superuser.is_superuser = True
        self.superuser.save()
        
    def test_business_exchange_email(self):
        """
        Tests sending an exchange to a business with no project.
        """
        self.client.login(username='superuser', password='test')
        exchange = self.create_exchange()
        person = self.create_person()
        self.create_relationship(data={
            'from_contact': exchange.business, 
            'to_contact': person
        })
        email_exchange_url = reverse('email_exchange', args=(exchange.pk,))
        response = self.client.get(email_exchange_url)
        self.assertEqual(response.status_code, 200)
        data = {
            'to': person.pk,
            'memo': self.random_string(500),
        }
        response = self.client.post(email_exchange_url, data)
        self.assertNoFormErrors(response)
        view_business_url = reverse('view_business', 
                                    args=(exchange.business.pk,))
        self.assertRedirects(response, view_business_url)
        self.assertEqual(len(mail.outbox), 1)
        message = mail.outbox[0]
        self.assertEqual(
            message.subject, 
            ' '.join([
                exchange.business.name, 
                exchange.type.label,
                datetime.datetime.today().strftime('%Y-%m-%d'),
            ]),
        )
        self.assertEqual(message.body.strip(), data['memo'])
        self.assertTrue(person.email in message.to)
        
    def test_project_exchange_email(self):
        """
        Tests sending an exchange to a business with no project.
        """
        self.client.login(username='superuser', password='test')
        exchange = self.create_exchange()
        person = self.create_person()
        self.create_relationship(data={
            'from_contact': exchange.business, 
            'to_contact': person
        })
        project = self.create_project(data={'business': exchange.business})
        self.create_project_relationship(data={
            'project': project,
            'contact': person,
        })
        transaction = self.create_transaction(data={'exchange': exchange,
                                                    'project': project,})
        email_exchange_url = reverse('email_exchange', args=(exchange.pk,))
        response = self.client.get(email_exchange_url)
        self.assertEqual(response.status_code, 200)
        data = {
            'to': person.pk,
            'memo': self.random_string(500),
        }
        response = self.client.post(email_exchange_url, data)
        self.assertNoFormErrors(response)
        view_business_url = reverse('view_business', 
                                    args=(exchange.business.pk,))
        self.assertRedirects(response, view_business_url)
        self.assertEqual(len(mail.outbox), 1)
        message = mail.outbox[0]
        self.assertEqual(
            message.subject, 
            ' '.join([
                project.name, 
                exchange.type.label,
                datetime.datetime.today().strftime('%Y-%m-%d'),
            ]),
        )
        self.assertEqual(message.body.strip(), data['memo'])
        self.assertTrue(person.email in message.to)



from setuptools import setup, find_packages

setup(
    name='minibooks',
    version='0.2.0',
    author='Caktus Consulting Group',
    author_email='solutions@caktusgroup.com',
    packages=find_packages(),
    include_package_data=True,
    exclude_package_data={
        '': ['*.sql', '*.pyc', 'example_project'],
    },
    url='http://bitbucket.org/caktus/minibooks',
    license='LICENSE.txt',
    description='minibooks is a simple CRM, bookkeeping and double entry accounting package for the Django web framework.',
    long_description=open('README.rst').read(),
    classifiers=[
        'Topic :: Internet :: WWW/HTTP :: Dynamic Content',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: BSD License',
        'Programming Language :: Python',
        'Topic :: Software Development :: Libraries :: Python Modules',
        'Development Status :: 4 - Beta',
        'Operating System :: OS Independent',
    ],
)




# -*- coding: utf-8 -*-
# ----------------------------------------------------------------------------
# $Id: management.py 372 2009-01-13 16:21:01Z copelco $
# ----------------------------------------------------------------------------
#
#    Copyright (C) 2008 Caktus Consulting Group, LLC
#
#    This file is part of minibooks.
#
#    minibooks is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Affero General Public License as
#    published by the Free Software Foundation, either version 3 of 
#    the License, or (at your option) any later version.
#    
#    minibooks is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU Affero General Public License for more details.
#    
#    You should have received a copy of the GNU Affero General Public License
#    along with minibooks.  If not, see <http://www.gnu.org/licenses/>.
#


from django.dispatch import dispatcher
from django.db.models import signals
from django.contrib.auth.models import User, Group

from crm import models as crm_models

from minibooks.ledger import demo
from minibooks.ledger import models as ledger_models


def post_syncdb(signal, sender, app, created_models, **kwargs):
    # only run when our model got created
    if (signal == signals.post_syncdb) and (app == ledger_models):
        ledger_models.install()
    if (signal == signals.post_syncdb) and (app == crm_models):
        crm_models.install()

signals.post_syncdb.connect(post_syncdb)


def clean_minibooks_db():
    for g in Group.objects.all():
        g.delete()
    for u in User.objects.all():
        u.delete()
    for b in crm_models.Business.objects.all():
        b.delete()
    for p in crm_models.Project.objects.all():
        p.delete()
    
    demo.setup_demo()



# -*- coding: utf-8 -*-
# ----------------------------------------------------------------------------
# $Id$
# ----------------------------------------------------------------------------
#
#    Copyright (C) 2008 Caktus Consulting Group, LLC
#
#    This file is part of minibooks.
#
#    minibooks is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Affero General Public License as
#    published by the Free Software Foundation, either version 3 of 
#    the License, or (at your option) any later version.
#    
#    minibooks is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU Affero General Public License for more details.
#    
#    You should have received a copy of the GNU Affero General Public License
#    along with minibooks.  If not, see <http://www.gnu.org/licenses/>.
#


from django.conf.urls.defaults import *

import minibooks.ledger.views as views

urlpatterns = patterns('',
    url(r'^cron/', views.cron, name='cron'),
    
    # exchange
    url(r'^exchange/list/$', views.list_exchanges, name='list_exchanges'),
    url(r'^exchange/list/(?P<exchange_type_slug>[A-za-z-]+)/$', views.list_exchanges, name='list_exchanges_by_type'),
    url(r'^(?:business/(?P<business_id>\d+)/)?(?:project/(?P<project_id>\d+)/)?exchange/create/(?P<exchange_type_slug>[A-za-z-]+)/$', views.create_edit_exchange, name='create_exchange'),
    url(r'^exchange/(?P<exchange_id>\d+)/pdf/$', views.view_exchange_as_pdf, name='view_exchange_as_pdf'),
    url(r'^exchange/(?P<exchange_id>\d+)/edit/(?P<exchange_type_slug>[A-za-z-]+)/$', views.create_edit_exchange, name='edit_exchange'),
    url(
        r'^exchange/(?P<exchange_id>\d+)/edit/type/(?P<exchange_type_slug>[A-za-z-]+)/$',
        views.edit_exchange_type,
        name='edit_exchange_type',
    ),
    url(r'^exchange/(?P<exchange_id>\d+)/remove/$', views.remove_exchange, name='remove_exchange'),
    url(r'^exchange/(?P<exchange_id>\d+)/duplicate/$', views.duplicate_exchange, name='duplicate_exchange'),
    
    # transaction
    url(r'^transaction/reconcile/$', views.reconcile_transaction, name='reconcile_transaction'),
    
    # report
    url(r'^project/report/(?P<project_id>\d+)/$', views.show_project_report, name='show_project_report'),
    
    # emails
    url(r'^new/email/exchange/(?P<exchange_id>\d+)/$', views.new_exchange_email, name='email_exchange'),
    url(r'^new/email/project_report/(?P<project_id>\d+)/$', views.new_project_report_email, name='email_project_report'),
    
    # account
    url(r'^account/list/$', views.list_accounts, name='list_accounts'),
    url(r'^account/transfer/$', views.transfer_funds, name='transfer_funds'),
    url(r'^account/transfer/(?P<transaction_id>\d+)/$', views.transfer_funds, name='edit_account_transfer'),
    
    url(
        r'^account/general-entry/$',
        views.general_entry,
        name='create_general_entry',
    ),
    url(
        r'^account/general-entry/(?P<transaction_id>\d+)/$',
        views.general_entry,
        name='edit_general_entry',
    ),
    url(r'^account/transfer/debit/list/$', views.list_transfer_debit_accounts, name='list_transfer_debit_accounts'),
    url(r'^account/print/$', views.account_ledger, name='account_ledger'),
    url(
        r'^account/(?P<account_id>\d+)/$',
        views.account_ledger,
        name='single_account_ledger',
    ),
    url(
        r'^account/(?P<account_id>\d+)/business-balances/$',
        views.account_business_balances,
        name='account_business_balances',
    ),
    url(
        r'^account/(?P<account_id>\d+)/remove/$',
        views.remove_account,
        name='remove_account',
    ),
    url(
        r'^account/(?P<account_id>\d+)/edit/$',
        views.create_edit_account,
        name='edit_account',
    ),
    url(
        r'^account/create/$',
        views.create_edit_account,
        name='create_account',
    ),
    
    url(
        r'^report/$',
        views.list_reports,
        name='list_reports',
    ),
    url(
        r'^report/profit-loss/$',
        views.profit_loss,
        name='profit_loss',
    ),
    url(
        r'^report/balance-sheet/$',
        views.balance_sheet,
        name='balance_sheet',
    ),
)



import datetime

from django.contrib.auth.models import User

from crm import models as crm
from minibooks.ledger import models as ledger

def setup_demo():
    try:
        administrator = User.objects.get(username='demo')
    except User.DoesNotExist:
        administrator = User.objects.create_user(
            username='demo',
            email='demo@minibooksdemo.com',
            password='demo',
        )
        administrator.first_name = 'Demo'
        administrator.last_name = 'User'
        administrator.is_staff = True
        administrator.is_superuser = True
        administrator.save()
        crm.Profile.objects.create(user=administrator)
    
    people = (
        ('jdoe', 'John', 'Doe', 'johndoe@hometownanimals.com'),
        ('bjohnson', 'Bill', 'Johnson', 'bill@anicecreamshop.com'),
    )
    for username, first_name, last_name, email in people:
        user = User(
            username=username,
            first_name=first_name,
            last_name=last_name,
            email=email,
        )
        user.set_unusable_password()
        user.save()
        profile = crm.Profile.objects.create(user=user)
    
    businesses = (
        ('Home Town Animal Shelter',
            'Small animal shelter in the area',
            'jdoe',
            'Client',
            '444444 Franklin St.',
            'Chapel Hill',
            'NC',
            '27510',
        ),
        ('An Ice Cream Shop',
            'Local ice cream store with homemade flavors!',
            'bjohnson',
            'Client',
            '222222 Rosemary St.',
            'Chapel Hill',
            'NC',
            '27510',
        ),
        ('Coffee Shop', 
            'Where we take clients for a chat',
            None,
            'Vendor',
            None,
            None,
            None,
            None,
        ),
    )
    for name, notes, contact, type, street, city, state, zipcode in businesses:
        if street:
            address = crm.Address.objects.create(
                street=street,
                city=city,
                state=state,
                zip=zipcode,
            )
        else:
            address = None
        business = crm.Business.objects.create(
            name=name,
            notes=notes,
        )
        business.business_types.add(crm.BusinessType.objects.get(name=type))
        if address:
            business.address = address
            business.save()
        if contact:
            business.contacts.add(User.objects.get(username=contact))
    
    jdoe = User.objects.get(username='jdoe')
    bjohnson = User.objects.get(username='bjohnson')
    ice_cream_shop = crm.Business.objects.get(name='An Ice Cream Shop')
    animal_shelter = crm.Business.objects.get(name='Home Town Animal Shelter')
    coffee_shop = crm.Business.objects.get(name='Coffee Shop')
    
    ice_cream_website = crm.Project.objects.create(
        name='Ice Cream Shop website',
        type='software',
        status='accepted',
        description='Simple CMS website',
        business=ice_cream_shop,
        point_person=administrator,
    )
    crm.ProjectRelationship.objects.create(
        user=bjohnson,
        project=ice_cream_website,
    )
    crm.ProjectRelationship.objects.create(
        user=administrator,
        project=ice_cream_website,
    )
    
    animal_shelter_website = crm.Project.objects.create(
        name='Animal Shelter website',
        type='software',
        status='finished',
        description='Custom database driven website with animal listings',
        business=animal_shelter,
        point_person=administrator,
    )
    crm.ProjectRelationship.objects.create(
        user=jdoe,
        project=animal_shelter_website,
    )
    crm.ProjectRelationship.objects.create(
        user=administrator,
        project=animal_shelter_website,
    )
    
    consulting = ledger.Account.objects.get(name='Consulting')
    ar = ledger.Account.objects.get(name='Accounts Receivable')
    checking = ledger.Account.objects.get(name='Checking')
    general_expenses = ledger.Account.objects.get(name='General Expenses')
    
    invoice = ledger.ExchangeType.objects.get(label='Invoice')
    receipt = ledger.ExchangeType.objects.get(label='Receipt')
    purchase = ledger.ExchangeType.objects.get(label='Purchase')
    
    exchange = ledger.Exchange.objects.create(
        memo='Invoice for redesign and implementation of website',
        date=datetime.date(2008, 1, 1),
        date_due=datetime.date(2008, 1, 15),
        type=invoice,
        business=ice_cream_shop,
    )
    exchange.transactions.create(
        date=datetime.date(2008, 1, 1),
        debit=ar,
        credit=consulting,
        project=ice_cream_website,
        memo='Implement simple Django CMS',
        amount='55.0',
        quantity='10',
    )
    exchange.transactions.create(
        date=datetime.date(2008, 1, 1),
        debit=ar,
        credit=consulting,
        project=ice_cream_website,
        memo='Design website',
        amount='75.0',
        quantity='7',
    )
    exchange = ledger.Exchange.objects.create(
        memo='partial payment for website',
        date=datetime.date(2008, 1, 14),
        type=receipt,
        business=ice_cream_shop,
    )
    exchange.transactions.create(
        date=datetime.date(2008, 1, 1),
        debit=checking,
        credit=ar,
        project=ice_cream_website,
        memo='partial payment for website',
        amount='600.0',
        quantity='1.0',
    )
    
    exchange = ledger.Exchange.objects.create(
        memo='Invoice for custom database solution',
        date=datetime.date(2008, 2, 20),
        date_due=datetime.date(2008, 3, 10),
        type=invoice,
        business=animal_shelter,
    )
    exchange.transactions.create(
        date=datetime.date(2008, 2, 5),
        debit=ar,
        credit=consulting,
        project=animal_shelter_website,
        memo='Implement simple Django CMS',
        amount='55.0',
        quantity='10',
    )
    exchange.transactions.create(
        date=datetime.date(2008, 2, 8),
        debit=ar,
        credit=consulting,
        project=animal_shelter_website,
        memo='Animal listing model and interface',
        amount='125.0',
        quantity='25',
    )
    exchange.transactions.create(
        date=datetime.date(2008, 2, 16),
        debit=ar,
        credit=consulting,
        project=animal_shelter_website,
        memo='Patient tracking and appointment calendar',
        amount='125.0',
        quantity='45',
    )
    exchange.transactions.create(
        date=datetime.date(2008, 2, 1),
        debit=ar,
        credit=consulting,
        project=animal_shelter_website,
        memo='Layout and design',
        amount='75.0',
        quantity='10',
    )
    exchange = ledger.Exchange.objects.create(
        memo='partial payment for website',
        date=datetime.date(2008, 2, 25),
        type=receipt,
        business=animal_shelter,
    )
    exchange.transactions.create(
        date=datetime.date(2008, 2, 25),
        debit=checking,
        credit=ar,
        project=animal_shelter_website,
        memo='partial payment for website',
        amount='3000.0',
        quantity='1.0',
    )
    exchange = ledger.Exchange.objects.create(
        memo='partial payment for website',
        date=datetime.date(2008, 3, 1),
        type=receipt,
        business=animal_shelter,
    )
    exchange.transactions.create(
        date=datetime.date(2008, 3, 1),
        debit=checking,
        credit=ar,
        project=animal_shelter_website,
        memo='partial payment for website',
        amount='3000.0',
        quantity='1.0',
    )
    exchange = ledger.Exchange.objects.create(
        memo='partial payment for website',
        date=datetime.date(2008, 3, 10),
        type=receipt,
        business=animal_shelter,
    )
    exchange.transactions.create(
        date=datetime.date(2008, 3, 10),
        debit=checking,
        credit=ar,
        project=animal_shelter_website,
        memo='partial payment for website',
        amount='3000.0',
        quantity='1.0',
    )
    
    
    exchange = ledger.Exchange.objects.create(
        memo='met with potential client',
        date=datetime.date(2008, 3, 10),
        type=purchase,
        business=coffee_shop,
    )
    exchange.transactions.create(
        date=datetime.date(2008, 4, 2),
        debit=general_expenses,
        credit=checking,
        memo='cofee and lunch',
        amount='23.50',
        quantity='1.0',
    )
    
    jdoe.interactions.create(
        date=datetime.date(2008, 2, 14),
        type='phone',
        completed=True,
        project=animal_shelter_website,
        memo='Phone discussion on interface elements',
    )
    
    